import matplotlib
# Force matplotlib to not use any Xwindows backend.
matplotlib.use('Agg')
rcParams = matplotlib.rcParams
rcParams['svg.fonttype'] = 'none' # No text as paths. Assume font installed.

rcParams['font.serif'] = ['Times New Roman']
rcParams['font.sans-serif'] = ['Helvetica']
rcParams['font.family'] = 'sans-serif'

import numpy
import pylab
import PIL
import os

from PIL import Image
import mahotas
import pyslic

import scipy
import scipy.misc

import time
from datetime import timedelta

if not os.path.exists('results'):
	os.mkdir('results')

lowquality = ['c6-3.tif', 'c8-1.tif', 'c8-4.tif', 'c9-1.tif', 'c9-3.tif', 'd10-1.tif', 'd10-3.tif', 'd3-3.tif', 'd3-4.tif', 'd4-3.tif', 'd4-4.tif', 'd4-5.tif', 'd5-1.tif', 'd5-4.tif', 'd6-1.tif', 'd8-3.tif', 'd8-4.tif', 'd8-5.tif', 'e10-5.tif', 'e4-5.tif', 'e5-2.tif', 'e6-4.tif', 'e7-3.tif', 'e8-1.tif', 'e9-3.tif', 'e9-4.tif', 'f3-2.tif', 'f3-3.tif']
lowquality.append('c7-1.tif')
lowquality.append('c10-1.tif')
lowquality.append('d6-4.tif')
lowquality.append('d10-2.tif')
lowquality.append('e6-3.tif')


newfiles = [x for x in os.listdir('.') if '.tif' in x[-4:] and 'test' not in x and x not in lowquality]
timestamps = []
EXP_START_TIME = time.mktime(time.strptime('1:30:00 PM', '%I:%M:%S %p'))
for x in newfiles:
	tt = open(x.replace('.tif', '.txt')).next()
	timestamps.append(time.mktime(time.strptime(tt.split('\t')[2], '%I:%M:%S %p'))-EXP_START_TIME)


from skimage.exposure import equalize_adapthist, equalize_hist
# for x,t in newfiles:
# 	q = Image.open(x)
# 	accq = []
# 	for i in range(3):
# 		q.seek(i)
# 		z = numpy.array(q)
# 		accq.append(z)
# 	#
# 	blah = numpy.zeros((512, 512, 3), dtype='uint16')
# 	blah[:,:,1] = accq[1]
# 	blah[:,:,2] = mahotas.stretch(accq[0],0,65535, dtype='uint16')
# 	blah[:,:,1] = blah[:,:,1]*equalize_adapthist(blah[:,:,1], ntiles_x=4, ntiles_y=4)
# 	# blah[:,:,1] = skimage.exposure.adjust_log(blah[:,:,1], inv=True)
# 	for i in [1,2]:
# 		blah[:,:,i] = mahotas.stretch(blah[:,:,i], 0, 2**16-1, dtype='uint16')
# 	scipy.misc.imsave('aha-'+x+'.png', blah)

def fieldfeatures(green, blue):
	image = pyslic.Image()
	image.channels[ 'protein' ] = 1
	image.channels[ 'dna' ] = 0
	image.channeldata['protein'] = mahotas.stretch(green, 0,65535, dtype='uint16')
	image.channeldata['dna'] = mahotas.stretch(blue, 0,65535, dtype='uint16')
	image.loaded = True
	features = pyslic.computefeatures(image, 'field-dna+')
	return features

newfiles = sorted(zip(newfiles, timestamps), key=lambda x:x[1])

if not os.path.exists('rawfeats.npy'):
	accimages = []
	for x,t in newfiles:
		q = Image.open(x)
		accq = []
		for i in range(3):
			q.seek(i)
			z = numpy.array(q)
			accq.append(z)
			scipy.misc.imsave(x+'ch'+str(i)+'.png', q)
		blah = numpy.zeros((512, 512, 3), dtype='uint16')
		lowv = numpy.percentile(accq[1],5)
		# highv = numpy.percentile(accq[1],98)
		accq[1][accq[1]<=lowv]=lowv
		# accq[1][accq[1]>=highv]=highv
		blah[:,:,2] = accq[0]
		blah[:,:,1] = accq[1]
		# blah[:,:,1] = equalize_hist(blah[:,:,1])
		for i in [1,2]:
			blah[:,:,i] = mahotas.stretch(blah[:,:,i], 0, 2**16-1, dtype='uint16')
		scipy.misc.imsave('comb-'+x+'.png', blah)
		accimages.append(blah)

	accfeats = []
	for x in accimages:
		accfeats.append(fieldfeatures(x[:,:,2], x[:,:,1]))

	rawfeats = numpy.vstack(accfeats)
	numpy.save('rawfeats.npy',rawfeats)
else:
	rawfeats = numpy.load('rawfeats.npy')


# well, drug
platemap = dict(zip([item for sublist in [[x+str(y) for y in range(3,11)] for x in 'cdef'] for item in sublist],[16,4,48]*(10*4)))

# econazole, cycloheximide, vehicle
dmap = { 16 : 0, 4: 1, 48: 2}

# order: drug, drug, control (Josh)
design = numpy.zeros((rawfeats.shape[0],5))
uniq = {}
for ii,(f,ts) in enumerate(newfiles):
	w,r = f.replace('.tif','').split('-')
	if w not in uniq:
		uniq[w] = len(uniq)
	pl = int(w[1:])
	design[ii,3] = ts
	design[ii,4] = dmap[platemap[w]]
	design[ii,0] = int(r)
	design[ii,1] = uniq[w]

# for ii,(a,b,c) in enumerate([(y[0][0],int(y[0][1:]),y[1]) for y in [(x[0].split('.')[0].split('-')[0],x[1]) for x in newfiles]]):
# 	design[ii,(b-1)%3] = 1
# 	design[ii,3] = c
# 	design[ii,4] = (b-1)%3

# numpy.vstack([numpy.std(rawfeats[design[:,4]==z],axis=0) for z in range(3)])/numpy.std(rawfeats,axis=0)

X = rawfeats-numpy.mean(rawfeats,axis=0)
X /= numpy.std(X,axis=0)

from scipy.spatial.distance import cdist

# wsel = []
# for ii,(x,y,z) in enumerate(sorted([(x, y, numpy.mean(design[design[:,1]==y,3])) for (x,y) in uniq.iteritems()], key=lambda z:z[-1])):
# 	asub = X[design[:,1]==y]
# 	D = numpy.argmin(numpy.max(cdist(asub,asub),axis=1))
# 	wsel.append(numpy.where(design[:,1]==y)[0][D])

# wsel = numpy.array(wsel)

def linearly_independent_subset(V, threshold=1.e-2, return_orthogonal_basis=False):
    V = numpy.array(V, copy=True)
    orthogonal = []
    used = []
    for i,u in enumerate(V):
        for v in orthogonal:
            u -= numpy.dot(u,v)/numpy.dot(v,v) * v
        if numpy.dot(u,u) > threshold:
            orthogonal.append(u)
            used.append(i)
    if return_orthogonal_basis:
        return numpy.array(used),numpy.array(orthogonal)
    return numpy.array(used)

# okfeats = linearly_independent_subset(subX)
# X = X.copy()
X = X[:,linearly_independent_subset(X)]

vehicle = numpy.where(design[:,4]==2)[0]
econazole = numpy.where(design[:,4]==0)[0]
cycloheximide = numpy.where(design[:,4]==1)[0]

# print 1/0



import itertools
import heapq

def minimax_cluster(X):
        level = [[x] for x in range(X.shape[0])]
        bests = { x : x for x in range(X.shape[0]) }
        alive = range(X.shape[0])
        alives = []
        pairdists = []
        decisions = []
        from scipy.spatial.distance import cdist
        for a,b in itertools.combinations(alive,2):
                together = level[a]+level[b]
                D = numpy.max(cdist(X[together,:],X[together]),axis=1)
                mm = numpy.argmin(D)
                heapq.heappush(pairdists,(D[mm], (together[mm],a,b)))
        #
        toppoint = X.shape[0]
        Z = []
        while len(alive)>1:
                v, atop = heapq.heappop(pairdists)
                while ( atop[1] not in alive ) or ( atop[2] not in alive ):
                        v, atop = heapq.heappop(pairdists)
                print len(alive), len(level)
                alive = [x for x in alive if x not in [atop[1], atop[2]]]
                newish = level[atop[1]]+level[atop[2]]
                level.append(newish)
                for a in alive:
                        together = level[a]+newish
                        D = numpy.max(cdist(X[together,:],X[together]),axis=1)
                        mm = numpy.argmin(D)
                        heapq.heappush(pairdists,(D[mm], (together[mm],a, toppoint)))
                alive.append(toppoint)
                alives.append(alive)
                decisions.append((atop[1], atop[2]))
                bests[toppoint] = atop[0]
                Z.append([atop[1], atop[2], v, len(newish)])
                toppoint += 1
        #
        Z = numpy.asarray(Z)
        return Z, alives, level, bests


# find images with single nuclei to aid comparison
from skimage.filters.rank import otsu
from skimage.morphology import disk
from skimage.filters import gaussian_filter
from skimage.morphology import label
num_nuclei = numpy.zeros((len(newfiles),))
for ii,x in enumerate(newfiles):
	v = numpy.array(Image.open('comb-'+x[0]+'.png'))
	q = gaussian_filter(v[:,:,2], 8)
	q *= 255/numpy.max(q)
	q = q.astype('uint')
	num_nuclei[ii] = numpy.max(label(q>=mahotas.otsu(q)))

# manual 
num_nuclei[[ii for (ii,x) in enumerate(newfiles) if x[0]=='c10-2.tif']] = 2
single_nuclei = numpy.where(num_nuclei==1)[0]

Z, alives, level, bests = minimax_cluster(X[single_nuclei])

for i in range(1, len(alives)):
	ok = [x for x in alives[-i] if len(level[x])>1]
	if len(ok)==5:
		break

# ok = [x for x in alives[-7]]
rep_of_well = single_nuclei[numpy.array([bests[x] for x in ok])]
# most_central = single_nuclei[bests[alives[-1][0]]]


rep_of_well = numpy.array(sorted(rep_of_well,key=lambda x: design[x,4]))
rep_images = [numpy.array(Image.open('comb-'+newfiles[x][0]+'.png')) for x in rep_of_well]



v = numpy.array(Image.open('comb-d9-2.tif.png'))
q = gaussian_filter(v[:,:,2], 8)
q *= 255/numpy.max(q)
q = q.astype('uint')
nucmask = label(q>=mahotas.otsu(q))

without_nuc = numpy.zeros(v.shape, v.dtype)
without_nuc[:,:,1] = v[:,:,1]*(1-nucmask)
without_nuc[:,:,2] = v[:,:,2]
scipy.misc.imsave('example-without.png', without_nuc)

only_nuc = numpy.zeros(v.shape, v.dtype)
only_nuc[:,:,1] = v[:,:,1]*nucmask
only_nuc[:,:,2] = v[:,:,2]
scipy.misc.imsave('example-only.png', only_nuc)

only_nuc = numpy.zeros(v.shape, v.dtype)
only_nuc[:,:,0] = v[:,:,1]*nucmask
only_nuc[:,:,2] = v[:,:,2]
scipy.misc.imsave('example-only-red.png', only_nuc)


print 1/0









def makenice(xs):
	margin = numpy.ones((xs[0].shape[0], 20, xs[0].shape[2]),dtype=xs[0].dtype)*numpy.iinfo(xs[0].dtype).max
	# the one-liner is unintelligble because python is a stupid language
	acc = []
	for x in xs[:-1]:
		acc.append(x)
		acc.append(margin)
	acc.append(xs[-1])
	return numpy.hstack(acc)

repstacked = makenice(rep_images)

scipy.misc.imsave('reps.png', repstacked)
print 'making reps'

#not_reps = numpy.array(sorted(set(range(X.shape[0]))-set(rep_of_well)))
not_reps = numpy.array(sorted(set(single_nuclei)-set(rep_of_well)))

not_rep_cond = design[not_reps,4]
rep_cond = design[rep_of_well,4]

# if False:
# 	def farthest_first(xs,nn=3):
# 		startx = X[xs]
# 		inside=numpy.arange(startx.shape[0])==numpy.argmin(numpy.sum((startx-numpy.mean(startx,axis=0))**2,axis=1))
# 		# inside = numpy.arange(startx.shape[0])==[numpy.argmin(numpy.median(cdist(startx,startx),axis=1))]
# 		for i in range(nn):
# 			inside[numpy.where(~inside)[0][numpy.argmax(numpy.min(cdist(startx[~inside], startx[inside]),axis=1))]]=True
# 		return xs[inside]

# 	# vehicle_reps = farthest_first(vehicle)
# 	# econazole_reps = farthest_first(econazole)
# 	# cycloheximide_reps = farthest_first(cycloheximide)

# 	# rep_of_well = numpy.hstack([vehicle_reps, econazole_reps, cycloheximide_reps])
# 	rep_of_well = farthest_first(numpy.arange(X.shape[0]),12)

# 	# ffX = numpy.zeros((X.shape[1], X.shape[0]), order='F')
# 	# ffX[:] = X.T

# 	# Z = spams.archetypalAnalysis(ffX)
# 	# rep_of_well = numpy.argsort(cdist(Z.T, X),axis=1)[:,0]

# 	rep_of_well = numpy.array(sorted(rep_of_well,key=lambda x: design[x,4]))
# 	rep_images = [numpy.array(Image.open('comb-'+newfiles[x][0]+'.png')) for x in rep_of_well]

# 	scipy.misc.imsave('reps.png', numpy.hstack(rep_images))

# 	not_reps = numpy.array(sorted(set(range(X.shape[0]))-set(rep_of_well)))

import spams
fX = numpy.zeros((X.shape[1], not_reps.shape[0]),order='F')
dA = numpy.zeros((X.shape[1], rep_of_well.shape[0]),order='F')
fX[:] = X[not_reps].T
dA[:] = X[rep_of_well].T

not_rep_cond = design[not_reps,4]
rep_cond = design[rep_of_well,4]

print 'spams lasso'
modalpha_train = numpy.asarray(spams.lasso(fX,mode=0, D=dA,lambda1=1, verbose=10).todense().T)
print numpy.sum((fX-numpy.dot(dA, modalpha_train.T))**2)

print numpy.max(numpy.sum(numpy.abs(modalpha_train),axis=1))
print numpy.sort(numpy.sum(modalpha_train==0,axis=1))
print numpy.bincount(numpy.sum(modalpha_train!=0,axis=1))

as_percents = numpy.abs(modalpha_train)
as_percents = as_percents/numpy.sum(as_percents,axis=1)[:,numpy.newaxis]



from sklearn.linear_model import LogisticRegression, LogisticRegressionCV
LR = LogisticRegressionCV(cv=3, multi_class='multinomial')
LR.fit(modalpha_train,not_rep_cond)
pred_vehicle = LR.predict(modalpha_train)==not_rep_cond
print numpy.mean(pred_vehicle)
print 'accuracy against all three', numpy.mean(pred_vehicle), 'P=', scipy.stats.distributions.binom.sf(numpy.sum(pred_vehicle), pred_vehicle.shape[0], numpy.max(numpy.bincount(not_rep_cond.astype('int')))/float(pred_vehicle.shape[0]))

away_from = LR.predict_proba(modalpha_train)
print 'best cases', [newfiles[not_reps[x]][0] for x in [numpy.where(not_rep_cond==z)[0][numpy.argmax(away_from[not_rep_cond==z][:,0])] for z in range(3)]]




Z = numpy.vstack([numpy.dot(modalpha_train, LR.coef_[0,:]), numpy.dot(modalpha_train, LR.coef_[1,:])]).T

# Z = numpy.hstack([numpy.dot(modalpha_train,cyclo_or_econ_proj.T)+cyclo_or_econ_int, numpy.dot(modalpha_train,is_vehicle_proj.T)+is_vehicle_int])
mycolors = ['#1b9e77','#d95f02', '#7570b3']
mysymbs = ['^', 's', 'o']
for i in range(3):
	f = pylab.plot(Z[not_rep_cond==i,0], Z[not_rep_cond==i,1], mysymbs[i], color=mycolors[i], markersize=4, markeredgecolor='none')
	v, w = numpy.linalg.eigh(numpy.cov(Z[not_rep_cond==i,:], rowvar=0))
	u = w[0] / numpy.linalg.norm(w[0])
	angle = numpy.arctan(u[1] / u[0])
	angle = 180 * angle / numpy.pi  # convert to degrees
	ell = matplotlib.patches.Ellipse(numpy.mean(Z[not_rep_cond==i,:], axis=0), 2*numpy.sqrt(5.991*v[0]), 2*numpy.sqrt(5.991*v[1]), 180 + angle, edgecolor=mycolors[i], fill=False)
	#ell.set_alpha(0.5)
	pylab.gca().add_artist(ell)

# pylab.plot(Z[design[wsel,4]==0,0], Z[design[wsel,4]==0,1], '-', color=mycolors[0])

pylab.savefig('results/within_lasso_proj.svg')
pylab.close()



neg_rep_images = []
for i in rep_images:
	x = i.copy()
	x[:,:,0] = x[:,:,1]
	x[:,:,1] = 0
	neg_rep_images.append(x)

for i in numpy.where(not_rep_cond==0)[0]:
	v = numpy.array(Image.open('comb-'+newfiles[not_reps[i]][0]+'.png'))
	blah = [v]
	ratios = as_percents[i,:]/numpy.max(as_percents[i,:])
	for j in range(len(rep_images)):
		if modalpha_train[i,j]<0:
			blah.append(neg_rep_images[j]*ratios[j])
		else:
			blah.append(rep_images[j]*ratios[j])
	scipy.misc.imsave('results/econ-proj-'+newfiles[not_reps[i]][0]+'.png', makenice(blah))


for i in numpy.where(not_rep_cond==1)[0]:
	v = numpy.array(Image.open('comb-'+newfiles[not_reps[i]][0]+'.png'))
	blah = [v]
	ratios = as_percents[i,:]/numpy.max(as_percents[i,:])
	for j in range(len(rep_images)):
		if modalpha_train[i,j]<0:
			blah.append(neg_rep_images[j]*ratios[j])
		else:
			blah.append(rep_images[j]*ratios[j])
	scipy.misc.imsave('results/cyclo-proj-'+newfiles[not_reps[i]][0]+'.png', makenice(blah))

for i in numpy.where(not_rep_cond==2)[0]:
	v = numpy.array(Image.open('comb-'+newfiles[not_reps[i]][0]+'.png'))
	blah = [v]
	ratios = as_percents[i,:]/numpy.max(as_percents[i,:])
	for j in range(len(rep_images)):
		if modalpha_train[i,j]<0:
			blah.append(neg_rep_images[j]*ratios[j])
		else:
			blah.append(rep_images[j]*ratios[j])
	scipy.misc.imsave('results/veh-proj-'+newfiles[not_reps[i]][0]+'.png', makenice(blah))




from sklearn.linear_model import LogisticRegression, LogisticRegressionCV

standardized_times = design[:,3] - numpy.mean(design[:,3])
standardized_times /= numpy.std(standardized_times)

idxn = numpy.arange(X.shape[0])#[design[wsel,4] != 1]
LR = LogisticRegressionCV(cv=3, solver='liblinear')
classes = (design[idxn,4]==2).astype('int')
LR.fit(X[idxn,:],classes)
pred_vehicle = LR.predict(X[idxn,:])==classes
print numpy.mean(pred_vehicle)
is_vehicle_proj = LR.coef_.copy()
is_vehicle_int = LR.intercept_

print 'accuracy against vehicle', numpy.mean(pred_vehicle), 'P=', scipy.stats.distributions.binom.sf(numpy.sum(pred_vehicle), pred_vehicle.shape[0], numpy.max(numpy.bincount(classes))/float(pred_vehicle.shape[0]))


idxn = design[:,4] != 2
LR = LogisticRegressionCV(cv=3, solver='liblinear')
classes = (design[idxn,4]==0).astype('int')
LR.fit(X[idxn,:],classes)
pred_drug = LR.predict(X[idxn,:])==classes
print numpy.mean(pred_drug)
cyclo_or_econ_proj = LR.coef_.copy()
cyclo_or_econ_int = LR.intercept_
print 'accuracy between drugs', numpy.mean(pred_drug), 'P=', scipy.stats.distributions.binom.sf(numpy.sum(pred_drug), pred_drug.shape[0], numpy.max(numpy.bincount(classes))/float(pred_drug.shape[0]))

Z = numpy.hstack([numpy.dot(X,is_vehicle_proj.T)+is_vehicle_int, numpy.dot(X,cyclo_or_econ_proj.T)+cyclo_or_econ_int])
mycolors = ['#1b9e77','#d95f02', '#7570b3']
mysymbs = ['^', 's', 'o']
for i in range(3):
	f = pylab.plot(Z[design[:,4]==i,0], Z[design[:,4]==i,1], mysymbs[i], color=mycolors[i], markersize=4, markeredgecolor='none')
	v, w = numpy.linalg.eigh(numpy.cov(Z[design[:,4]==i,:], rowvar=0))
	u = w[0] / numpy.linalg.norm(w[0])
	angle = numpy.arctan(u[1] / u[0])
	angle = 180 * angle / numpy.pi  # convert to degrees
	ell = matplotlib.patches.Ellipse(numpy.mean(Z[design[:,4]==i,:], axis=0), 2*numpy.sqrt(5.991*v[0]), 2*numpy.sqrt(5.991*v[1]), 180 + angle, edgecolor=mycolors[i], fill=False)
	#ell.set_alpha(0.5)
	pylab.gca().add_artist(ell)

# pylab.plot(Z[design[wsel,4]==0,0], Z[design[wsel,4]==0,1], '-', color=mycolors[0])
# pylab.ylim([-6, 6])
pylab.gca().spines['top'].set_visible(False)
pylab.gca().spines['right'].set_visible(False)
pylab.gca().xaxis.set_ticks_position('bottom')
pylab.gca().tick_params(axis='x', direction='out')
pylab.gca().tick_params(axis='y', direction='out')
pylab.gca().yaxis.set_ticks_position('left')
pylab.ylim([-5, 4])
pylab.savefig('results/separator.svg')
pylab.close()

