#!/home/awn/tools/bin/python
#import cv2
import numpy
import sys
import os
import string
import scipy.stats
import random
import ic100
import scipy
import scipy.misc
import scipy.ndimage
import itertools
import scipy.spatial

os.chdir('/home/twoactive/filtering')

def kdist(bw,axs,ays):
    if ays.shape[0] < axs.shape[0]:
        ys = axs
        xs = ays
    else:
        ys = ays
        xs = axs        
    acc = 0.0
    inv_sq_bw = -0.5/bw**2
    dd = numpy.zeros(ys.shape)
    for i in range(xs.shape[0]):
        dd[:] = ys[:]
        dd -= xs[i,:]
        acc += numpy.sum(numpy.exp(inv_sq_bw*numpy.sum(dd**2,axis=1)))
    return acc/(xs.shape[0]*ys.shape[0])

def fullmmdl(bw,xs,ys):
    return kdist(bw,xs,xs)+kdist(bw,ys,ys)-2.0*kdist(bw,xs,ys)


def rawmmdl(bw,xs,ys):
    touched_xs = set([])
    touched_ys = set([])
    acc = 0.0
    inv_sq_bw = 1.0/bw**2
    maxiter = min(ys.shape[0],xs.shape[0])/2
    if maxiter > 1:
        for ii in range(maxiter):
                x1 = numpy.random.randint(0,xs.shape[0])
                while x1 in touched_xs:
                        x1 = numpy.random.randint(0,xs.shape[0])
                touched_xs.add(x1)
                x2 = numpy.random.randint(0,xs.shape[0])
                while x2 in touched_xs:
                        x2 = numpy.random.randint(0,xs.shape[0])
                touched_xs.add(x2)
                y1 = numpy.random.randint(0,ys.shape[0])
                while y1 in touched_ys:
                        y1 = numpy.random.randint(0,ys.shape[0])
                touched_ys.add(y1)
                y2 = numpy.random.randint(0,ys.shape[0])
                while y2 in touched_ys:
                        y2 = numpy.random.randint(0,ys.shape[0])
                touched_ys.add(y2)
                acc += numpy.exp(-0.5*numpy.sum((xs[x1,:]-xs[x2,:])**2)*inv_sq_bw)
                acc += numpy.exp(-0.5*numpy.sum((ys[y1,:]-ys[y2,:])**2)*inv_sq_bw)
                acc -= numpy.exp(-0.5*numpy.sum((xs[x1,:]-ys[y2,:])**2)*inv_sq_bw)
                acc -= numpy.exp(-0.5*numpy.sum((xs[x2,:]-ys[y1,:])**2)*inv_sq_bw)
        return acc/maxiter
    else:
        return numpy.nan

def mmdl(bw,xs,ys):
    z = rawmmdl(bw,xs,ys)
    while z < 0:
        z = rawmmdl(bw,xs,ys)
    return z

markedbad = []
for f in open('baddies.txt','r'):
    x = f.rstrip().replace('.png', '.npy')
    markedbad.append('filter/surfcache/'+x)

markedgood = []
for f in open('goodies.txt','r'):
    x = f.rstrip().replace('.png', '.npy')
    markedgood.append('filter/surfcache/'+x)

probablywtf = set(markedgood) & set(markedbad)

print 'probablywtf should be empty', probablywtf

datagood = [numpy.load(x) for x in markedgood if x not in probablywtf]
databad = [numpy.load(x) for x in markedbad if x not in probablywtf]
databad = [x for x in databad if x.shape[0] > 5]
ngoods = len(datagood)
nbads = len(databad)

possible_bws = numpy.logspace(numpy.log(0.001), numpy.log(10), 300, base=numpy.e)

worder = range(possible_bws.shape[0])
numpy.random.shuffle(worder)

gg_idxn = numpy.zeros((ngoods*(ngoods-1)/2,2),dtype='int')
gg_dist = numpy.zeros((gg_idxn.shape[0],))
bb_idxn = numpy.zeros((nbads*(nbads-1)/2,2),dtype='int')
bb_dist = numpy.zeros((bb_idxn.shape[0],))
gb_idxn = numpy.zeros((ngoods*nbads,2),dtype='int')
gb_dist = numpy.zeros((gb_idxn.shape[0],))
ii = 0
for g1,g2 in itertools.combinations(range(ngoods),2):
    gg_idxn[ii,0] = g1
    gg_idxn[ii,1] = g2
    ii += 1

ii = 0
for b1,b2 in itertools.combinations(range(nbads),2):
    bb_idxn[ii,0] = b1
    bb_idxn[ii,1] = b2
    ii += 1

ii = 0
for gi,bi in itertools.product(range(ngoods),range(nbads)):
    gb_idxn[ii,0] = gi
    gb_idxn[ii,1] = bi
    ii += 1


NFOLDS = 5
total_gplaces = numpy.arange(ngoods)
total_bplaces = numpy.arange(nbads)
numpy.random.shuffle(total_gplaces)
numpy.random.shuffle(total_bplaces)
gfolds = numpy.array_split(total_gplaces, NFOLDS)
bfolds = numpy.array_split(total_bplaces, NFOLDS)

needed_gg = numpy.zeros(gg_idxn.shape[0],dtype=bool)
for i in range(NFOLDS):
    needed_gg[numpy.logical_xor(numpy.in1d(gg_idxn[:,0], gfolds[i]),numpy.in1d(gg_idxn[:,1], gfolds[i]))] = True

needed_gb = numpy.zeros(gb_idxn.shape[0],dtype=bool)
for i in range(NFOLDS):
    needed_gb[numpy.logical_xor(numpy.in1d(gb_idxn[:,0], gfolds[i]),numpy.in1d(gb_idxn[:,1], bfolds[i]))] = True

needed_bb = numpy.zeros(bb_idxn.shape[0],dtype=bool)
for i in range(NFOLDS):
    needed_bb[numpy.logical_xor(numpy.in1d(bb_idxn[:,0], bfolds[i]),numpy.in1d(bb_idxn[:,1], bfolds[i]))] = True


(numpy.sum(needed_gg)+numpy.sum(needed_gb)+numpy.sum(needed_bb))/(0.0+needed_gg.shape[0]+needed_gb.shape[0]+needed_bb.shape[0])





for wbandwidth in worder:
    if not os.path.exists('bandwidth-'+str(wbandwidth)+'.csv'):
        fout = open('bandwidth-'+str(wbandwidth)+'.csv','w')
        bw = possible_bws[wbandwidth]



        ii = 0
        for g1,g2 in itertools.combinations(range(ngoods),2):
            if numpy.mod(ii,100)==0:
                print ii
            gg_idxn[ii,0] = g1
            gg_idxn[ii,1] = g2
            gg_dist[ii] = mmdl(bw, datagood[g1],datagood[g2])
            ii += 1
        #
        ii = 0
        for b1,b2 in itertools.combinations(range(nbads),2):
            if numpy.mod(ii,100)==0:
                print ii
            bb_idxn[ii,0] = b1
            bb_idxn[ii,1] = b2
            bb_dist[ii] = mmdl(bw, databad[b1],databad[b2])
            ii += 1
        #
        ii = 0
        for gi,bi in itertools.product(range(ngoods),range(nbads)):
            if numpy.mod(ii,100)==0:
                print ii
            gb_idxn[ii,0] = gi
            gb_idxn[ii,1] = bi
            gb_dist[ii] = mmdl(bw, datagood[gi],databad[bi])
            ii += 1
        # cross validate
        gperf = numpy.zeros((len(datagood),))
        for gi in range(len(datagood)):
            gd_left = gg_dist[gg_idxn[:,0]==gi]
            gd_right = gg_dist[gg_idxn[:,1]==gi]
            bd = gb_dist[gb_idxn[:,0]==gi]
            gdm = numpy.nanmin(numpy.hstack([gd_left,gd_right]))
            bdm = numpy.nanmin(bd)
            gperf[gi] = gdm - bdm
        #
        fout.write(' '.join([str(x) for x in gperf])+"\n")
        bperf = numpy.zeros((len(databad),))
        for bi in range(len(databad)):
            bd_left = bb_dist[bb_idxn[:,0]==bi]
            bd_right = bb_dist[bb_idxn[:,1]==bi]
            gd = gb_dist[gb_idxn[:,1]==bi]
            gdm = numpy.nanmin(gd)
            bdm = numpy.nanmin(numpy.hstack([bd_left,bd_right]))
            bperf[bi] = bdm-gdm
        fout.write(' '.join([str(x) for x in bperf])+"\n")
        fout.close()

