# pre:  the WELLMAP files (stored in /data), which were produced by the robot experiment code
#       "justfeats", which are a snapshot of the features we computed in parallel to the image directories
#       bad_images.csv, which was produced by the filtering script
# post: This script produces "calculated/zscored_data.npy" and the parallel "calculated/indices.npy" files along the way
#       to calculated/REPORT-[round].csv, which store summary data for the accuracy measure for figure 2, etc.

import numpy
import sys
import os
import string
import scipy.stats
import scipy.spatial.distance

# lifted from Luis's code in order to not import mahotas and all the rest of that elephant
def linearly_independent_subset(V, threshold=1.e-5, return_orthogonal_basis=False):
    V = numpy.array(V, copy=True)
    orthogonal = []
    used = []
    for i,u in enumerate(V):
        for v in orthogonal:
            u -= numpy.dot(u,v)/numpy.dot(v,v) * v
        if numpy.dot(u,u) > threshold:
            orthogonal.append(u)
            used.append(i)
    if return_orthogonal_basis:
        return numpy.array(used),numpy.array(orthogonal)
    return numpy.array(used)


if not os.path.exists('calculated/rawdata.npy'):
    print 'building rawdata'
    badfeats = set([])
    for rline in open('data/bad_images.csv','r'):
        Line = string.split(rline.rstrip(), '/')
        badfeats.add(Line[-1])
    #
    well_to_experiment = {}
    for z in os.listdir('data'):
        if 'WELLMAP' in z:
            partnames = string.split(z[:-4], '-')
            thename = partnames[1]
            f = open('data/'+z,'r')
            next(f)
            for rline in f:
                Line = string.split(rline.rstrip())
                well_to_experiment[(thename, Line[0], int(Line[1]))] = (int(Line[4]), int(Line[2]))
    #
    experiment_to_fields = {}
    for z in os.listdir('justfeats'):
        Line = [x for x in string.split(z,'_') if len(x)>0]
        if z not in badfeats:
            if 'batch' in Line[0]:
                thename = 'round'+Line[0][5:]
            else:
                thename = Line[0]
            place = well_to_experiment[(thename, Line[1], int(Line[2]))]
            if place not in experiment_to_fields:
                experiment_to_fields[place] = []
            experiment_to_fields[place].append('justfeats/'+z)
    #
    thedata = numpy.zeros((numpy.sum([len(x) for x in experiment_to_fields.values()]), 173))
    theindices = numpy.zeros((thedata.shape[0],2), dtype='i2')
    i = 0
    for e,l in experiment_to_fields.iteritems():
        for f in l:
            theindices[i,0] = e[0]
            theindices[i,1] = e[1]
            thedata[i,:] = numpy.load(f)
            i += 1
    #
    numpy.save('calculated/rawdata.npy', thedata)
    numpy.save('calculated/indices.npy', theindices)
else:
    print 'loading indices'
    theindices = numpy.load('calculated/indices.npy')
    if not os.path.exists('calculated/zscored_data.npy'):
        print 'building zscored data'
        thedata = numpy.load('calculated/rawdata.npy')
        mean_of_thedata = numpy.mean(thedata,axis=0)
        thedata -= mean_of_thedata
        std_of_thedata = numpy.std(thedata,axis=0)
        thedata /= std_of_thedata
        #
        numpy.save('calculated/rawdata_mean.npy', mean_of_thedata)
        numpy.save('calculated/rawdata_std.npy', std_of_thedata)
        thedata = thedata[:,numpy.where(numpy.logical_not(numpy.isnan(numpy.std(thedata,axis=0))))[0]]
        #
        goodfeats = linearly_independent_subset(thedata.T)
        thedata = thedata[:,goodfeats]
        numpy.save('calculated/zscored_data.npy', thedata)
    else:
        print 'loading zscored data'
        thedata = numpy.load('calculated/zscored_data.npy')

print 'building exp_to_range'
exp_to_range = {}
for i in range(theindices.shape[0]):
    if (theindices[i,0], theindices[i,1]) not in exp_to_range:
        exp_to_range[(theindices[i,0], theindices[i,1])] = []
    exp_to_range[(theindices[i,0], theindices[i,1])].append(i)

for i in exp_to_range.keys():
    j = exp_to_range[i]
    exp_to_range[i] = (numpy.min(j), numpy.max(j)+1)



def maxcount(arr):
    acc = {}
    for i in arr:
        if i not in acc:
            acc[i] = 0
        acc[i] += 1
    maxi = 0
    maxv = 0
    for i,v in acc.iteritems():
        if v > maxv:
            maxv = v
            maxi = i
    return maxi

def per_round_workload(which_round):
    if os.path.exists('calculated/REPORT-'+str(which_round)+'.csv'):
        return
    ap = open('sample/bear_round'+str(which_round)+'.apredictions','r')
    observed_experiments = []
    frontier_experiments = []
    exp_to_pheno = {}
    pheno_to_predicted = {}
    pheno_to_obs_exps = {}
    phenos = set([])
    predicted_experiments = []
    for rline in ap:
        Line = string.split(rline.rstrip())
        place = (int(Line[0]), int(Line[1]))
        if '1' in Line[4]:
            observed_experiments.append(place)
            if int(Line[2]) not in pheno_to_obs_exps:
                pheno_to_obs_exps[int(Line[2])] = []
            pheno_to_obs_exps[int(Line[2])].append(place)
        else:
            if int(Line[2]) not in pheno_to_predicted:
                pheno_to_predicted[int(Line[2])] = []
            pheno_to_predicted[int(Line[2])].append(place)
            if place in exp_to_range:
                predicted_experiments.append(place)
        if '1' in Line[5]:
            frontier_experiments.append(place)
        exp_to_pheno[place] = int(Line[2])
        phenos.add(int(Line[2]))
    #
    phenos = list(phenos)
    phenos_array = numpy.asarray(phenos)
    unobserved_calls = {}
    badphenos = []
    for ph in phenos:
        print "in phenotype", ph, 'round', which_round
        myname = 'calculated/COMPUTED-round-'+str(which_round)+'-ph-'+str(ph)+'.npy'
        if not os.path.exists(myname):
            # build the data for the phenotype
            field_ranges = []
            nk = 0
            for e in pheno_to_obs_exps[ph]:
                if e not in exp_to_range:
                    bads = open('WARNING-MISSING-DATA.txt','a+')
                    bads.write(str(which_round)+' '+str(e[0])+' '+str(e[1])+"\n")
                    print "WARNING: MISSING DATA FOR ", e
                    bads.flush()
                    bads.close()
                else:
                    field_ranges.append(exp_to_range[e])
                    nk += exp_to_range[e][1] - exp_to_range[e][0]
            ph_data = numpy.zeros((nk, thedata.shape[1]))
            i = 0
            for a,b in field_ranges:
                ph_data[i:(i+b-a),:] = thedata[a:b,:]
                i += b-a
            numpy.save(myname, ph_data)
            print "saved data"
        else:
            print "loading data"
            ph_data = numpy.load(myname)
        if ph_data.shape[0] == 0:
            bads = open('WARNING-MISSING-PHENOTYPE-DATA.txt','a+')
            bads.write(str(which_round)+' '+str(ph)+"\n")
            print "WARNING: MISSING DATA FOR PHENOTYPE", ph
            bads.flush()
            bads.close()
            badphenos.append(ph)
        else:
            # against = numpy.zeros(ph_data.shape)
            # for e in predicted_experiments:
            #     acc = []
            #     for f in range(exp_to_range[e][0], exp_to_range[e][1]):
            #         against = ph_data - thedata[f,:]
            #         acc.append(numpy.min(numpy.sum(against**2,axis=1)))
            #     if e not in unobserved_calls:
            #         unobserved_calls[e] = {}
            #     unobserved_calls[e][ph] = acc
            for e in predicted_experiments:
                if e not in unobserved_calls:
                    unobserved_calls[e] = {}
                # what is the smallest distance to phenotype-labeled points for each point in this experiment
                unobserved_calls[e][ph] = numpy.min(scipy.spatial.distance.cdist(thedata[range(exp_to_range[e][0], exp_to_range[e][1]),:], ph_data, metric='sqeuclidean'),axis=1)
    #
    if len(badphenos) > 0:
        print "WARNING: Discarding some phenotypes", badphenos
        newphenos = []
        for i in phenos:
            if i not in badphenos:
                newphenos.append(i)
        phenos = newphenos
    print 'computing accuracies', 'round', which_round
    accuracy_voting = 0
    frontier_accuracy_voting = 0
    accuracy_proportions = []
    pred_yes = []
    pred_no = []
    for pl in predicted_experiments:
        ph_to_mindists = unobserved_calls[pl]
        dmat = numpy.zeros((len(phenos), len(ph_to_mindists[phenos[0]])))
        for i in range(len(phenos)):
            dmat[i,:] = ph_to_mindists[phenos[i]]
        did_we_get_it_right = int(maxcount(phenos_array[numpy.argmin(dmat,axis=0)])==exp_to_pheno[pl])
        if pl in frontier_experiments:
            frontier_accuracy_voting += did_we_get_it_right
        accuracy_voting += did_we_get_it_right
        accuracy_proportions.append(1.0*numpy.where(phenos_array[numpy.argmin(dmat,axis=0)]==exp_to_pheno[pl])[0].shape[0]/dmat.shape[1])
        #
        dmat_prime = numpy.roll(dmat, phenos_array.shape[0] - numpy.where(phenos_array==exp_to_pheno[pl])[0],axis=0)
        choices = numpy.random.randint(1,dmat.shape[0],size=dmat.shape[1])
        pred_no.extend(list(dmat_prime[choices, numpy.arange(dmat.shape[1])]))
        pred_yes.extend(list(dmat_prime[0,:]))
    #
    print 'saving report', 'round', which_round
    outs = open('calculated/REPORT-'+str(which_round)+'.csv','w')
    outs.write(' '.join([str(x) for x in [which_round, accuracy_voting, len(accuracy_proportions), 
            numpy.min(accuracy_proportions), numpy.percentile(accuracy_proportions,25), 
            numpy.percentile(accuracy_proportions,50), numpy.percentile(accuracy_proportions,75), numpy.max(accuracy_proportions),
            numpy.mean(accuracy_proportions), scipy.stats.mannwhitneyu(pred_yes, pred_no)[1], 1.0/len(pred_yes), 
            frontier_accuracy_voting]]))
    outs.close()


for which_round in range(1,31):
    print 'working on round', which_round
    per_round_workload(which_round)

# which_round = int(sys.argv[1])
# per_round_workload(which_round)
