import numpy
import sys
import os
import string
import scipy.stats

def linearly_independent_subset(V, threshold=1.e-5, return_orthogonal_basis=False):
    V = numpy.array(V, copy=True)
    orthogonal = []
    used = []
    for i,u in enumerate(V):
        for v in orthogonal:
            u -= numpy.dot(u,v)/numpy.dot(v,v) * v
        if numpy.dot(u,u) > threshold:
            orthogonal.append(u)
            used.append(i)
    if return_orthogonal_basis:
        return numpy.array(used),numpy.array(orthogonal)
    return numpy.array(used)

print 'loading marking data'
# find the ones that we additionally marked as 'bad'
markings = [x for x in os.listdir('filtering') if ( '.marking' in x ) and (x[0] == 'd')]
filtered_at_all = set([])
badfeats = set([])
for m in markings:
    for line in open('filtering/'+m,'r'):
        ss = line.rstrip().split()
        fname = ss[0].split('/')[-1]+'.npy'
        if float(ss[-1])==0:
            filtered_at_all.add(fname)
            if (float(ss[1])-float(ss[2])) > 0:
                badfeats.add(fname)

for line in open('filtering/verybad.txt'):
    badfeats.add(line.rstrip()[:-4]+'.npy')

skipped = 0
if not os.path.exists('calculated/rawdata_posthoc.npy'):
    print 'building rawdata'
    #badfeats = set([])
    for rline in open('data/bad_images.csv','r'):
        Line = string.split(rline.rstrip(), '/')
        badfeats.add(Line[-1])
    #
    # reconstruct file paths
    rpath = {}
    for line in open('sample/FULLDATA_field_info.txt','r'):
        ll = line.rstrip()
        rpath[ll.split('/')[-1]] = ll
    #
    well_to_experiment = {}
    for z in os.listdir('data'):
        if 'WELLMAP' in z:
            partnames = string.split(z[:-4], '-')
            thename = partnames[1]
            f = open('data/'+z,'r')
            next(f)
            for rline in f:
                Line = string.split(rline.rstrip())
                well_to_experiment[(thename, Line[0], int(Line[1]))] = (int(Line[4]), int(Line[2]))
    #
    experiment_to_fields = {}
    for z in os.listdir('justfeats'):
        Line = [x for x in string.split(z,'_') if len(x)>0]
        if z not in filtered_at_all:
            continue
        if z not in badfeats:
            if 'batch' in Line[0]:
                thename = 'round'+Line[0][5:]
            else:
                thename = Line[0]
            place = well_to_experiment[(thename, Line[1], int(Line[2]))]
            if place not in experiment_to_fields:
                experiment_to_fields[place] = []
            if z[:-4] in rpath: # keep only the ones we've filtered
                experiment_to_fields[place].append('justfeats/'+z)
        else:
            skipped += 1
    #
    thedata = numpy.zeros((numpy.sum([len(x) for x in experiment_to_fields.values()]), 173))
    theindices = numpy.zeros((thedata.shape[0],2), dtype='i2')
    i = 0
    field_order = []
    for e,l in experiment_to_fields.iteritems():
        for f in l:
            theindices[i,0] = e[0]
            theindices[i,1] = e[1]
            thedata[i,:] = numpy.load(f)
            field_order.append(f)
            i += 1
    #
    # for some reason, there are NaNs in these data...04/02/14
    ppx = numpy.logical_not(numpy.sum(numpy.isnan(thedata),axis=1))
    thedata = thedata[ppx,:]
    theindices = theindices[ppx,:]
    field_order = [field_order[x] for x in numpy.where(ppx)[0]]
    numpy.save('calculated/rawdata_posthoc.npy', thedata)
    numpy.save('calculated/indices_posthoc.npy', theindices)
    fout = open('calculated/posthoc_field_order.txt','w')
    for f in field_order:
        fout.write(rpath[f.split('/')[-1][:-4]]+"\n")
    fout.close()
else:
    print 'loading indices'
    theindices = numpy.load('calculated/indices_posthoc.npy')

# zz = numpy.load('indices.npy')
# here = set([frozenset(x) for x in theindices])
# there = set([frozenset(x) for x in zz])


if not os.path.exists('zscored_data_posthoc.npy'):
    print 'building zscored data'
    thedata = numpy.load('calculated/rawdata_posthoc.npy')
    mean_of_thedata = numpy.mean(thedata,axis=0)
    thedata -= mean_of_thedata
    std_of_thedata = numpy.std(thedata,axis=0)
    features_kept_so_far = numpy.where(std_of_thedata>0)[0]
    thedata = thedata[:,std_of_thedata>0]
    std_of_thedata = numpy.std(thedata,axis=0)
    thedata /= std_of_thedata
    #
    numpy.save('calculated/rawdata_mean_posthoc.npy', mean_of_thedata)
    numpy.save('calculated/rawdata_std_posthoc.npy', std_of_thedata)
    #thedata = thedata[:,numpy.where(numpy.logical_not(numpy.isnan(numpy.std(thedata,axis=0))))[0]]
    #
    goodfeats = linearly_independent_subset(thedata.T)
    features_kept_so_far = features_kept_so_far[goodfeats]
    numpy.save('calculated/features_kept_so_far.npy',features_kept_so_far)
    thedata = thedata[:,goodfeats]
    numpy.save('calculated/zscored_data_posthoc.npy', thedata)
else:
    print 'loading zscored data'
    thedata = numpy.load('calculated/zscored_data_posthoc.npy')
