import matplotlib
# Force matplotlib to not use any Xwindows backend.
matplotlib.use('Agg')

import numpy
import os
import pylab
import string
import itertools
import scipy
import scipy.stats
import scipy.spatial
import scipy.cluster
import scipy.cluster.hierarchy as sch

if not os.path.exists('posthoc_results'):
    os.mkdir('posthoc_results')

numpy.seterr(all='raise')


PREFIX = ''

NORMALIZE_INNER_BY_MAX = True

if NORMALIZE_INNER_BY_MAX:
    PREFIX += 'NORMALIZE_INNER_BY_MAX-'

#
# Helper functions
#

def linkage_to_control(Y):
    children_of = {}
    alive_at_level = [[]]
    for i in range(Y.shape[0]+1):
        children_of[i] = [i]
        alive_at_level[0].append(i)
    #
    for l in range(Y.shape[0]):
        alive_at_level.append([])
        i,j = Y[l,:2].astype('int')
        spot = Y.shape[0]+1+l
        children_of[spot] = children_of[i]+children_of[j]
        alive_at_level[l+1] = [x for x in alive_at_level[l] if x not in [i,j]]+[spot]
    #
    return children_of, alive_at_level

def myargsort(seq):
    # http://stackoverflow.com/questions/3382352/equivalent-of-numpy-argsort-in-basic-python/3383106#3383106
    #non-lambda version by Tony Veijalainen
    return [i for (v, i) in sorted((v, i) for (i, v) in enumerate(seq))]

#
# Load and arrange data
#

fin = open('metadata/drug_names.txt','r')
compound_labels = []
xx = 1
for line in fin:
    me = line.rstrip()
    if 'Alginate' not in line:
        me += 'mM'
    compound_labels.append(me)
    xx += 1

fin = open('metadata/clone_names.txt','r')
clone_labels = []
for line in fin:
    clone_labels.append(line.rstrip())

kosher_protein_list = [qwerty for qwerty in range(48) if clone_labels.index('NIH-3T3') != qwerty]

remapped_protein_subset = dict([(y+1,x) for (x,y) in enumerate(kosher_protein_list)])
fout = open('posthoc_results/remapped_clones.txt','w')
for x,y in remapped_protein_subset.items():
    fout.write(' '.join([str(x), str(y)])+"\n")

fout.close()

modified_clone_labels = []
for z, i in enumerate(clone_labels):
    if 'Chr' in i[:3]:
        modified_clone_labels.append('clone '+str(z+1))
    else:
        modified_clone_labels.append(i)

modified_clone_labels = [modified_clone_labels[x] for x in kosher_protein_list]

enumerated_clone_labels = [y+' ('+str(x+1).zfill(2)+')' for (x,y) in zip(kosher_protein_list,modified_clone_labels)]

data = numpy.load('calculated/zscored_data_posthoc.npy')
theindices = numpy.load('calculated/indices_posthoc.npy')

ranges_of_48 = {}
for i in range(theindices.shape[0]):
    e = ( theindices[i,0]-1 ) % 48 + 1, ( theindices[i,1]-1 ) % 48 + 1
    if e not in ranges_of_48:
        ranges_of_48[e] = []
    ranges_of_48[e].append(i)

imputed_experiments = []
for e in itertools.product(range(1,49),range(1,49)):
    if e not in ranges_of_48:
        ranges_of_48[e] = []
        imputed_experiments.append(e)
        print 'imputing data for experiment',e

linkorder48 = ranges_of_48.keys()
lookup48 = dict(zip(linkorder48, range(len(linkorder48))))

summary48 = numpy.zeros((48*48,data.shape[1]))
for e,i in lookup48.iteritems():
    if e not in imputed_experiments:
        summary48[i,:] = numpy.mean(data[ranges_of_48[e],:],axis=0)

# impute to the unperturbed data
for e in imputed_experiments:
    summary48[lookup48[e],:] = summary48[lookup48[(48,e[1])],:]

#
# this is the raw matrix of how much each drugged experiment differed in feature space from the corresponding unperturbed experiment
#
perturbations = numpy.zeros((47,47))
perturbations_imputed = numpy.zeros((47,47), dtype='bool')
for pp,p in enumerate(kosher_protein_list):
    for d in range(1,48):
        if (d,p+1) not in imputed_experiments:
            perturbations[pp,d-1] = numpy.sqrt(numpy.sum((summary48[lookup48[(d,p+1)],:]-summary48[lookup48[(48,p+1)],:])**2))
        else:
            perturbations_imputed[pp,d-1] = True


annotated_phenotypes = {}
xx = 1
for line in open('metadata/clone_assignments.txt','r'):
    ph = line.rstrip()
    if ph not in annotated_phenotypes:
        annotated_phenotypes[ph] = []
    annotated_phenotypes[ph].append(xx)
    xx += 1

for a in annotated_phenotypes.keys():
    b = annotated_phenotypes[a]
    border = numpy.argsort([clone_labels[x-1] for x in b])
    annotated_phenotypes[a] = [b[x] for x in border]



pheno_order = [
'nucleoli',
'nucleoli and cytoplasm',
'nucleoplasm', 
'nucleoplasm and cytoplasm',
'uniform cytoplasm',
'structured cytoplasm',
'focal',
'mitochondria',
'ER',
'ERGIC',
'Golgi',
'plasma membrane',
'vesicles',
'cytoskeleton']


basic_phenotype_groups = [[remapped_protein_subset[z] for z in annotated_phenotypes[x]] for x in pheno_order]
nbasic_phenotypes = len(basic_phenotype_groups)
groups_of_clones = [annotated_phenotypes[x] for x in pheno_order]

unaugmented_clone_order = reduce(lambda x,y: x+y,[[z for z in annotated_phenotypes[x]] for x in pheno_order])

# expsummary = numpy.zeros((nbasic_phenotypes*48,summary48.shape[1]))
# jj = 0
# phblocks = []
# drblocks = {}
# for ph, phg in enumerate(basic_phenotype_groups):
#     acc = []
#     zz = [numpy.vstack([summary48[lookup48[(d,p+1)],:] for p in phg if (d,p+1) in lookup48]) for d in range(1,49)]
#     for di,dv in enumerate(zz):
#         if di not in drblocks:
#             drblocks[di] = []
#         drblocks[di].append(jj)
#         acc.append(jj)
#         expsummary[jj,:] = numpy.mean(dv,axis=0)
#         jj += 1
#     phblocks.append(acc)




# import sklearn
# import sklearn.manifold
# import sklearn.metrics
# similarities = sklearn.metrics.euclidean_distances(expsummary-numpy.mean(expsummary,axis=0))
# mds = sklearn.manifold.MDS(n_components=2, max_iter=3000, eps=1e-9, random_state=1,dissimilarity="precomputed", n_jobs=1)
# pos = mds.fit(similarities).embedding_

# for z,pp in enumerate(drblocks.values()):
#     pylab.subplot(7,7,z+1)
#     pylab.plot(pos[:,0], pos[:,1], '.')
#     pylab.plot(pos[pp,0], pos[pp,1], 'r.')
#     pylab.axis('off')

# pylab.show()


# # save the groups of clones to a file for later processing
# fout = open('posthoc_results/groups_of_clones.txt','w')
# for g in groups_of_clones:
#     fout.write(' '.join([str(x) for x in g])+"\n")

# fout.close()

#
# normalize the perturbations matrix by phenotype
#

inner_perturbations = numpy.zeros((len(basic_phenotype_groups),perturbations.shape[1]))
unaugmented = numpy.zeros(perturbations.shape)
unaugmented_imputed = numpy.zeros(perturbations.shape, dtype='bool')
augmented = numpy.zeros((len(basic_phenotype_groups)+perturbations.shape[0],perturbations.shape[1]))
augmented_imputed = numpy.zeros((len(basic_phenotype_groups)+perturbations.shape[0],perturbations.shape[1]), dtype='bool')
augmented_rownames = []
unaugmented_rownames = []
header_row = []
jj = 0
zz = 0
augmented_clone_remap = {}
unaugmented_phenotype_groups = []
for ii,ph in enumerate(basic_phenotype_groups):
    max_in_group = numpy.max(perturbations[ph,:])
    inner_perturbations[ii,:] = numpy.mean(perturbations[ph,:],axis=0)
    if NORMALIZE_INNER_BY_MAX:
        inner_perturbations[ii,:] /= max_in_group
        augmented[jj,:] = inner_perturbations[ii,:]
    else:
        augmented[jj,:] = inner_perturbations[ii,:]
    augmented_rownames.append(pheno_order[ii])
    header_row.append(True)
    jj += 1
    for kk in range(len(ph)):
        augmented_clone_remap[ph[kk]] = jj+kk
    if NORMALIZE_INNER_BY_MAX:
        augmented[jj:(jj+len(ph)),:] = perturbations[ph,:]/max_in_group
    else:
        augmented[jj:(jj+len(ph)),:] = perturbations[ph,:]
    unaugmented[zz:(zz+len(ph)),:] = augmented[jj:(jj+len(ph)),:]
    unaugmented_phenotype_groups.append(range(zz,len(ph)+zz))
    augmented_imputed[jj:(jj+len(ph)),:] = perturbations_imputed[ph,:]
    unaugmented_imputed[zz:(zz+len(ph)),:] = perturbations_imputed[ph,:]
    augmented_rownames.extend([modified_clone_labels[z] for z in ph])
    unaugmented_rownames.extend([modified_clone_labels[z] for z in ph])
    header_row.extend([False]*len(ph))
    jj += len(ph)
    zz += len(ph)

numpy.save('posthoc_results/'+PREFIX+'unaugmented.npy', unaugmented)
fout = open('posthoc_results/'+PREFIX+'unaugmented-rowdata.txt','w')
for x,y in zip(unaugmented_clone_order, unaugmented_rownames):
    fout.write(str(x)+' '+y+"\n")

fout.close()


#
# Construct a clustering of the drugs
#

pdist_of_drugs = scipy.spatial.distance.pdist(inner_perturbations.T)

Z_drugs = scipy.cluster.hierarchy.linkage(pdist_of_drugs,method='average')
drugclus_children_of, drugclus_alive_at_level = linkage_to_control(Z_drugs)

CHOSEN_DRUG_CUTOFF = 30
scipy.cluster.hierarchy.dendrogram(Z_drugs, labels=compound_labels, color_threshold=Z_drugs[CHOSEN_DRUG_CUTOFF,2])
pylab.savefig('posthoc_results/'+PREFIX+'drug_clustering.svg', bbox_inches='tight')
pylab.clf(); pylab.close()

# drug_summary = numpy.zeros((47,nbasic_phenotypes))
# for ph, phg in enumerate(basic_phenotype_groups):
#     d = 48
#     meanph = numpy.mean(numpy.vstack([summary48[lookup48[(d,p+1)],:] for p in phg if (d,p+1) in lookup48]),axis=0)
#     for d in range(1,48):
#         here = numpy.mean(numpy.vstack([summary48[lookup48[(d,p+1)],:] for p in phg if (d,p+1) in lookup48]),axis=0)
#         drug_summary[d-1,ph] = numpy.sqrt(numpy.sum((meanph-here)**2))

# pdist_of_stuff = scipy.spatial.distance.pdist(drug_summary)
# Z_drugs = scipy.cluster.hierarchy.linkage(pdist_of_stuff,method='average')
# drugclus_children_of, drugclus_alive_at_level = linkage_to_control(Z_drugs)

# CHOSEN_DRUG_CUTOFF = 38
# scipy.cluster.hierarchy.dendrogram(Z_drugs, labels=compound_labels, color_threshold=Z_drugs[CHOSEN_DRUG_CUTOFF,2])
# pylab.savefig('posthoc_results/'+PREFIX+'drug_clustering.svg', bbox_inches='tight')
# pylab.clf(); pylab.close()

drug_groups = [drugclus_children_of[x] for x in drugclus_alive_at_level[CHOSEN_DRUG_CUTOFF]]
fout = open('posthoc_results/'+PREFIX+'groups_of_drugs.txt','w')
for g in drug_groups:
    fout.write(' '.join([str(x+1) for x in g])+"\n")

fout.close()

#
# ANOVA model
#

def fit_two_way_anova(perturbations, perturbations_imputed, basic_phenotype_groups,drug_groups):
    design = numpy.zeros((numpy.prod(perturbations.shape)-numpy.sum(perturbations_imputed), len(basic_phenotype_groups)+len(drug_groups)))
    response = numpy.zeros((design.shape[0],))
    i = 0
    for dd,dg in enumerate(drug_groups):
        for pp,phg in enumerate(basic_phenotype_groups):
            for d in dg:
                for p in phg:
                    if not perturbations_imputed[p,d]:
                        design[i,pp] = 1
                        design[i,nbasic_phenotypes+dd] = 1
                        response[i] = perturbations[p,d]
                        i += 1
    #
    ss, crap1, crap2, crap3 = numpy.linalg.lstsq(numpy.dot(design.T, design), numpy.dot(design.T, response))
    anova_fit = numpy.zeros(perturbations.shape)
    inner_anova_fit = numpy.zeros((len(basic_phenotype_groups),len(drug_groups)))
    for dd,dg in enumerate(drug_groups):
        inner_anova_fit[:,dd] = ss[nbasic_phenotypes+dd]
        anova_fit[:,dg] = ss[nbasic_phenotypes+dd]
    #
    for pp,phg in enumerate(basic_phenotype_groups):
        inner_anova_fit[pp,:] += ss[pp]
        anova_fit[phg,:] += ss[pp]
    #
    rss = numpy.sum((perturbations-anova_fit)**2)
    return rss, anova_fit, inner_anova_fit


rss,the_ANOVA_fit, the_ANOVA_fit_inner = fit_two_way_anova(unaugmented, unaugmented_imputed, unaugmented_phenotype_groups, drug_groups)


#
# Heatmap
#

def setup_dendrosides():
    # Compute and plot first dendrogram.
    fig = pylab.figure(figsize=(8,8))
    # left bottom width height
    ax1 = fig.add_axes([0.11,0.1,0.1,0.6])
    #Z1 = sch.dendrogram(Z_unperturbed, orientation='right', color_threshold=Z_unperturbed[UNPERTURBED_CUTOFF,2])
    #idx1 = Z1['leaves']
    ax1.set_xticks([])
    ax1.set_yticks([])
    ax1.set_frame_on(False) 
    # Compute and plot second dendrogram.
    # left bottom width height
    ax2 = fig.add_axes([0.3,0.89,0.6,0.1])
    Z2 = sch.dendrogram(Z_drugs, color_threshold=Z_drugs[CHOSEN_DRUG_CUTOFF,2])
    idx2 = Z2['leaves']
    ax2.set_xticks([])
    ax2.set_yticks([])
    ax2.set_frame_on(False) 
    return fig,idx2

def setup_heapmapper(fig,idx1,idx2, perturbations, row_names, column_names, markout, myname, boldentries=None):
    # Plot distance matrix.
    axmatrix = fig.add_axes([0.3,0.1,0.6,0.6*(perturbations.shape[0]/47.0)])
    # organize the data so that they are in the same order as the clusterings are
    subT = perturbations[idx1,:]
    subT = subT[:, idx2]
    xxx = [column_names[x] for x in idx2]
    yyy = [row_names[x] for x in idx1]
    for x in range(perturbations.shape[1]):
        for y in range(perturbations.shape[0]):
            if subT[y,x] >= 0:
                color1 = [1.0-subT[y,x],1.0-subT[y,x],1.0]
            else:
                color1 = [1.0,1.0+subT[y,x],1.0+subT[y,x]]
            axmatrix.fill([x,x,x+1,x+1], [y, y+1, y+1, y], fill=True, color=color1, edgecolor='white', linewidth=1)
    #
    # mark out imputed experiments (row, column)
    if markout is not None:
        fudgerino = 0.1
        markout2 = markout[idx1,:]
        markout2 = markout2[:,idx2]
        for x in range(perturbations.shape[1]):
            for y in range(perturbations.shape[0]):
                if markout2[y,x]:
                    axmatrix.plot([x+fudgerino,x+1-fudgerino],[y+fudgerino,y+1-fudgerino],'r-')
    #
    axmatrix.set_frame_on(False)
    # put the major ticks at the middle of each cell
    axmatrix.set_yticks(numpy.arange(perturbations.shape[0])+0.5, minor=False)
    axmatrix.set_xticks(numpy.arange(perturbations.shape[1])+0.5, minor=False)
    # want a more natural, table-like display
    axmatrix.invert_yaxis()
    axmatrix.xaxis.tick_top()
    axmatrix.set_aspect('auto')
    axmatrix.set_xlim([0,perturbations.shape[1]])
    axmatrix.set_ylim([0,perturbations.shape[0]])
    axmatrix.set_xticklabels(xxx, minor=False, fontsize=8, family='sans-serif') 
    axmatrix.set_yticklabels(yyy, minor=False, fontsize=8, family='sans-serif')
    pylab.xticks(rotation=90)
    axmatrix.grid(False)
    # Turn off all the ticks
    ax = pylab.gca()
    for t in axmatrix.xaxis.get_major_ticks(): 
        t.tick1On = False 
        t.tick2On = False
    #
    for t in axmatrix.yaxis.get_major_ticks(): 
        t.tick1On = False 
        t.tick2On = False 
    #
    if boldentries is not None:
        bbb = [boldentries[x] for x in idx1]
        [i[1].set_weight("bold") for i in zip(bbb,axmatrix.get_yticklabels()) if i[0]]
    pylab.savefig(myname, bbox_inches='tight')
    pylab.clf(); pylab.close()


idx1 = range(unaugmented.shape[0])[::-1]


fig,idx2 = setup_dendrosides()
if NORMALIZE_INNER_BY_MAX:
    setup_heapmapper(fig,idx1,idx2, unaugmented, unaugmented_rownames, compound_labels, unaugmented_imputed, 'posthoc_results/'+PREFIX+'biclustering.svg')
else:
    setup_heapmapper(fig,idx1,idx2, unaugmented/numpy.max(unaugmented), unaugmented_rownames, compound_labels, unaugmented_imputed, 'posthoc_results/'+PREFIX+'biclustering.svg')

fig,idx2 = setup_dendrosides()
setup_heapmapper(fig,idx1,idx2, the_ANOVA_fit, unaugmented_rownames, compound_labels, unaugmented_imputed, 'posthoc_results/'+PREFIX+'biclustering-anova-fit.svg')

delta = unaugmented-the_ANOVA_fit
fig,idx2 = setup_dendrosides()
setup_heapmapper(fig,idx1,idx2, delta, unaugmented_rownames, compound_labels, unaugmented_imputed, 'posthoc_results/'+PREFIX+'biclustering-delta-relative-to-anova-fit.svg')


delta = numpy.maximum(unaugmented-the_ANOVA_fit,0)
fig,idx2 = setup_dendrosides()
setup_heapmapper(fig,idx1,idx2, delta, unaugmented_rownames, compound_labels, unaugmented_imputed, 'posthoc_results/'+PREFIX+'biclustering-delta-over-anova-fit.svg')

# numpy.save('posthoc_results/'+PREFIX+'perturbations.npy', perturbations)
# numpy.save('posthoc_results/'+PREFIX+'the_ANOVA_fit.npy', the_ANOVA_fit)


fig = pylab.figure(figsize=(8,8))
# # left bottom width height
ax2 = fig.add_axes([0.3,1.07,0.6,0.1])
Z2 = sch.dendrogram(Z_drugs, color_threshold=Z_drugs[CHOSEN_DRUG_CUTOFF,2])
idx2 = Z2['leaves']
ax2.set_xticks([])
ax2.set_yticks([])
ax2.set_frame_on(False) 
idx1_augmented = range(augmented.shape[0])[::-1]
if NORMALIZE_INNER_BY_MAX:
    setup_heapmapper(fig,range(augmented.shape[0])[::-1],idx2, augmented, augmented_rownames, compound_labels, augmented_imputed, 'posthoc_results/'+PREFIX+'augmented.svg', boldentries=header_row)
else:
    setup_heapmapper(fig,range(augmented.shape[0])[::-1],idx2, augmented/numpy.max(augmented), augmented_rownames, compound_labels, augmented_imputed, 'posthoc_results/'+PREFIX+'augmented.svg', boldentries=header_row)


fig = pylab.figure(figsize=(8,8))
# # left bottom width height
ax2 = fig.add_axes([0.3,.48,0.6,0.1])
Z2 = sch.dendrogram(Z_drugs, color_threshold=Z_drugs[CHOSEN_DRUG_CUTOFF,2])
idx2 = Z2['leaves']
ax2.set_xticks([])
ax2.set_yticks([])
ax2.set_frame_on(False)
idx1_augmented = range(inner_perturbations.shape[0])[::-1]
if NORMALIZE_INNER_BY_MAX:
    setup_heapmapper(fig,idx1_augmented,idx2, inner_perturbations, pheno_order, compound_labels, None, 'posthoc_results/'+PREFIX+'inner.svg')
else:
    setup_heapmapper(fig,idx1_augmented,idx2, inner_perturbations/numpy.max(inner_perturbations), pheno_order, compound_labels, None, 'posthoc_results/'+PREFIX+'inner.svg')


inner_anova = numpy.zeros((the_ANOVA_fit_inner.shape[0],len(compound_labels)))
for i in range(inner_anova.shape[0]):
    for dd,dg in enumerate(drug_groups):
        inner_anova[i,dg] = the_ANOVA_fit_inner[i,dd]


fig = pylab.figure(figsize=(8,8))
# # left bottom width height
ax2 = fig.add_axes([0.3,.48,0.6,0.1])
Z2 = sch.dendrogram(Z_drugs, color_threshold=Z_drugs[CHOSEN_DRUG_CUTOFF,2])
idx2 = Z2['leaves']
ax2.set_xticks([])
ax2.set_yticks([])
ax2.set_frame_on(False) 
setup_heapmapper(fig,range(inner_anova.shape[0])[::-1],idx2, inner_anova, pheno_order, compound_labels, None, 'posthoc_results/'+PREFIX+'inner-anova-fit.svg')

fig = pylab.figure(figsize=(8,8))
# # left bottom width height
ax2 = fig.add_axes([0.3,.48,0.6,0.1])
Z2 = sch.dendrogram(Z_drugs, color_threshold=Z_drugs[CHOSEN_DRUG_CUTOFF,2])
idx2 = Z2['leaves']
ax2.set_xticks([])
ax2.set_yticks([])
ax2.set_frame_on(False) 
setup_heapmapper(fig,range(inner_anova.shape[0])[::-1],idx2, inner_perturbations-inner_anova, pheno_order, compound_labels, None, 'posthoc_results/'+PREFIX+'inner-delta-relative-to-anova-fit.svg')

fig = pylab.figure(figsize=(8,8))
# # left bottom width height
ax2 = fig.add_axes([0.3,.48,0.6,0.1])
Z2 = sch.dendrogram(Z_drugs, color_threshold=Z_drugs[CHOSEN_DRUG_CUTOFF,2])
idx2 = Z2['leaves']
ax2.set_xticks([])
ax2.set_yticks([])
ax2.set_frame_on(False) 
setup_heapmapper(fig,range(inner_anova.shape[0])[::-1],idx2, numpy.maximum(inner_perturbations-inner_anova,0), pheno_order, compound_labels, None, 'posthoc_results/'+PREFIX+'inner-delta-over-anova-fit.svg')



# zout = open('zout.csv','w')
# zout.write('clone condition pert'+"\n")
# for ppx,pp in enumerate(basic_phenotype_groups):
#     for ddx,dd in enumerate(drug_groups):
#         for p in pp:
#             for d in dd:
#                 zout.write(' '.join(['c'+str(ppx), 'd'+str(ddx), str(perturbations[p,d])])+"\n")

# # for p in range(perturbations.shape[0]):
# #     for d in range(perturbations.shape[1]):
# #         zout.write('"'+modified_clone_labels[p]+'" "'+compound_labels[d]+'" '+str(perturbations[p,d])+"\n")

# zout.close()

#
# Predictions vs Posthoc...
#


# pylab.hist(perturbations.flat,100)
# pylab.xlabel('Degree of Perturbation (arb units)')
# pylab.ylabel('Counts')
# pylab.savefig('posthoc_results/'+PREFIX+'perturbation_histogram.svg', bbox_inches='tight')
# pylab.clf(); pylab.close()


# a,b = scipy.cluster.vq.kmeans2(perturbations.flatten(),2, minit='points', iter=200)
# threshold_for_ground_truth = (a[0]+a[1])/2


pylab.hist(unaugmented.flat,100)
pylab.xlabel('Degree of Perturbation (arb units)')
pylab.ylabel('Counts')
pylab.savefig('posthoc_results/'+PREFIX+'perturbation_histogram.svg', bbox_inches='tight')
pylab.clf(); pylab.close()


# 20150219 scipy authors changed the code so that mininit doesn't work anymore :/
# a,b = scipy.cluster.vq.kmeans2(unaugmented.flatten(),2, minit='points', iter=200)
a,b = scipy.cluster.vq.kmeans2(unaugmented.flatten(),2, iter=200)
threshold_for_ground_truth = (a[0]+a[1])/2

fig,idx2 = setup_dendrosides()
setup_heapmapper(fig,idx1,idx2, unaugmented>=threshold_for_ground_truth, unaugmented_rownames, compound_labels, unaugmented_imputed, 'posthoc_results/'+PREFIX+'biclustering-thresholded.svg')

fig,idx2 = setup_dendrosides()
setup_heapmapper(fig,idx1,idx2, unaugmented>=threshold_for_ground_truth, unaugmented_rownames, compound_labels, unaugmented_imputed, 'posthoc_results/'+PREFIX+'biclustering-thresholded.png')
# from sklearn import mixture
# clf = mixture.GMM(n_components=2, covariance_type='full')
# clf.fit(unaugmented.flatten())
# zz = clf.predict_proba(numpy.linspace(0,1,50))
# pylab.plot(numpy.linspace(0,1,50), zz[:,0], 'k-')
# pylab.plot(numpy.linspace(0,1,50), zz[:,1], 'r-')
# pylab.show()

# fout = open('THINGS_FOR_PAPER.txt','a')
# fout.write('The AR(1) parameters are: '+' '.join([str(x[-1]) for x in [beta_total_null, beta_plate_null, beta_plate_and_quad_null, beta_null, beta]])+"\n")
# fout.write('There is some variance to these, and slight departures from what is in the paper should be irrelevant'+"\n")
# fout.close()

print numpy.sum(unaugmented.flat>threshold_for_ground_truth), float(numpy.sum(unaugmented.flat>threshold_for_ground_truth))/numpy.prod(unaugmented.shape)

exp_to_range = {}
for i in range(theindices.shape[0]):
    if (theindices[i,0], theindices[i,1]) not in exp_to_range:
        exp_to_range[(theindices[i,0], theindices[i,1])] = []
    exp_to_range[(theindices[i,0], theindices[i,1])].append(i)

#linkorder96_subset = [x for x in exp_to_range.keys() if ( (x[1] - 1) % 48 in kosher_protein_list ) and (48,x[1]) in exp_to_range]

linkorder96_subset = [x for x in exp_to_range.keys() if ( (x[1] - 1) % 48 +1 in unaugmented_clone_order ) and (48,x[1]) in exp_to_range]

linkorder96_subset_set = set(linkorder96_subset)

ap = open('sample/bear_round30.apredictions','r')
exp_to_pheno = {}
all_exp_to_pheno = {}
pheno_to_exp = {}
for rline in ap:
    Line = string.split(rline.rstrip())
    e = (int(Line[0]), int(Line[1]))
    all_exp_to_pheno[e] = int(Line[2])
    if e in linkorder96_subset_set:
        exp_to_pheno[e] = int(Line[2])
        if int(Line[2]) not in pheno_to_exp:
            pheno_to_exp[int(Line[2])] = []
        pheno_to_exp[int(Line[2])].append(e)


linkorder96_map = dict(zip(linkorder96_subset, range(len(linkorder96_subset))))

summary96 = numpy.zeros((len(linkorder96_subset),data.shape[1]))
for i,e in enumerate(linkorder96_subset):
    summary96[i,:] = numpy.mean(data[exp_to_range[e],:],axis=0)


pheno_to_representation = {}
for ph,es in pheno_to_exp.iteritems():
    places = []
    for e in [x for x in es if x in linkorder96_subset_set]:
        places.extend(exp_to_range[e])
    pheno_to_representation[ph] = numpy.mean(data[places,:],axis=0)

drugged_96 = [x for x in linkorder96_subset if (x[0] not in [48,96]) and exp_to_pheno[x] in pheno_to_representation]


ground_truth_distances = numpy.zeros((len(drugged_96),))
model_distances = numpy.zeros((len(drugged_96),))
model_differed = numpy.zeros((len(drugged_96),), dtype='bool')

for i,e in enumerate(drugged_96):
    #ground_truth_distances[i] = perturbations[remapped_protein_subset[(e[1]-1)%48+1],(e[0]-1)%48]
    ground_truth_distances[i] = unaugmented[unaugmented_clone_order.index((e[1]-1)%48+1), (e[0]-1)%48]
    model_distances[i] = numpy.sqrt(numpy.sum((pheno_to_representation[exp_to_pheno[e]]-summary96[linkorder96_map[(48,e[1])],:])**2))
    model_differed[i] = exp_to_pheno[e]!=exp_to_pheno[(48,e[1])]


truet = ground_truth_distances >= threshold_for_ground_truth
truef = numpy.logical_not(truet)

fout = open('THINGS_FOR_PAPER.txt','a')
fout.write('The posthoc filtered accuracy is '+str(float(numpy.sum(truet*model_differed+truef*(1-model_differed)))/len(drugged_96))+"\n")
fout.close()



unique_distances = numpy.unique(model_distances)[::-1]

TP = numpy.zeros((unique_distances.shape[0],))
FP = numpy.zeros((unique_distances.shape[0],))
TN = numpy.zeros((unique_distances.shape[0],))
FN = numpy.zeros((unique_distances.shape[0],))
for threshold in range(unique_distances.shape[0]):
    predt = model_distances >= unique_distances[threshold]
    predf = numpy.logical_not(predt)
    TP[threshold] = numpy.sum(numpy.logical_and(predt, truet))
    FP[threshold] = numpy.sum(numpy.logical_and(predt, truef))
    TN[threshold] = numpy.sum(numpy.logical_and(predf, truef))
    FN[threshold] = numpy.sum(numpy.logical_and(predf, truet))

fpr = FP/(FP+TN)
tpr = TP/(TP+FN)
precision = TP/(TP+FP)
recall = TP/(TP+FN)
pylab.figure()
pylab.plot(recall, precision, 'k.-'); 
pylab.xlim(0,1)
pylab.ylim(0,1)
pylab.ylabel('precision')
pylab.xlabel('recall')
pylab.savefig('posthoc_results/'+PREFIX+'_precision_recall.svg')
pylab.clf(); pylab.close()



myfpr = fpr/fpr[-1]
mytpr = tpr/tpr[-1]
if ( myfpr[0] != 0 ) or (mytpr[0] != 0):
    myfpr = numpy.hstack([[0], myfpr])
    mytpr = numpy.hstack([[0], mytpr])

auc = (numpy.trapz(y=tpr/tpr[-1], x=fpr/fpr[-1]))
pylab.figure()
pylab.plot(myfpr, mytpr ,'k.-')
pylab.ylabel('TPR')
pylab.xlabel('FPR')
pylab.title('AUC = '+"%.3f" % auc)
pylab.xlim(0,1)
pylab.ylim(0,1)
pylab.savefig('posthoc_results/'+PREFIX+'_ROC.svg')
pylab.clf(); pylab.close()

fout = open('THINGS_FOR_PAPER.txt','a')
fout.write('The posthoc filtered ROC AUC is '+str(auc)+"\n")
fout.close()



# 20150219 addition: of the quads that differed from the undrugged phenotype 4,3,2,1,0 times, how many are considered perturbed by
# "ground truth" models


# we want to handle the cases for phenotype 597, and 2323 gracefully since we threw out the data for them
# in the posthoc filtering





a,b = scipy.cluster.vq.kmeans2(perturbations.flatten(),2, iter=200)
threshold_for_global_ground_truth = (a[0]+a[1])/2



from_to_ph_comparisons = {}
model_confusions = numpy.zeros((5, 2, 2))
quad_confusion_to_global_ground_truth = numpy.zeros((5,2))
quad_confusion_to_local_truth = numpy.zeros((5,2))
obs_quad_confusion_to_global_ground_truth = numpy.zeros((5,2))
obs_quad_confusion_to_local_truth = numpy.zeros((5,2))
grand_local_confusion = numpy.zeros((2,2))
grand_global_confusion = numpy.zeros((2,2))
for d in range(1,47):
    for p in range(1,48):
        nominalph = all_exp_to_pheno[(48,p)]
        predphs = []
        obsphs = []
        for doff in [0,48]:
            for poff in [0,48]:
                predphs.append(all_exp_to_pheno[(d+doff,p+poff)])
                if (d+doff,p+poff) in exp_to_range:
                    obsphs.append(all_exp_to_pheno[(d+doff,p+poff)])
        ndiffs = 0
        nobsdiffs = numpy.sum([x != nominalph for x in obsphs])
        dacc = numpy.zeros((2,2))
        for pp in predphs:
            if ( nominalph, pp ) not in from_to_ph_comparisons:
                if nominalph in pheno_to_representation and pp in pheno_to_representation:
                    from_to_ph_comparisons[nominalph, pp] = numpy.sqrt(numpy.sum((pheno_to_representation[nominalph]-pheno_to_representation[pp])**2))
                else:
                    from_to_ph_comparisons[nominalph, pp] = numpy.infty
            ground_differed = from_to_ph_comparisons[nominalph, pp]>threshold_for_ground_truth
            pred_differed = pp != nominalph
            ndiffs += pred_differed
            dacc[int(pred_differed), int(ground_differed)] += 1
        model_confusions[ndiffs,:,:] += dacc
        #  don't include the untagged parental in this calculation
        if p != 46: 
            ldiff = int(unaugmented[unaugmented_clone_order.index(p), d-1]>threshold_for_ground_truth)
            gdiff = int(perturbations[kosher_protein_list.index(p-1),d-1]>threshold_for_global_ground_truth)
            quad_confusion_to_local_truth[ndiffs,ldiff]+=1 
            quad_confusion_to_global_ground_truth[ndiffs,gdiff]+=1 
            grand_local_confusion[0,ldiff] += 4-ndiffs
            grand_local_confusion[1,ldiff] += ndiffs
            grand_global_confusion[0,gdiff] += 4-ndiffs
            grand_global_confusion[1,gdiff] += ndiffs
            if len(obsphs)>0:
                obs_quad_confusion_to_local_truth[nobsdiffs,ldiff]+=1 
                obs_quad_confusion_to_global_ground_truth[nobsdiffs,gdiff]+=1 


# let's make a ROC curve, where we ask for varying numbers of OBSERVED experiments within a quad differing in phenotype from the 
# undrugged phenotype, how well we do compared to our "ground truth" model

tp = numpy.zeros((4,))
fp = numpy.zeros((4,))
tn = numpy.zeros((4,))
fn = numpy.zeros((4,))
for thresh in range(1,5):
    tp[thresh-1] = numpy.sum(obs_quad_confusion_to_local_truth[thresh:,1])
    fp[thresh-1] = numpy.sum(obs_quad_confusion_to_local_truth[thresh:,0])
    tn[thresh-1] = numpy.sum(obs_quad_confusion_to_local_truth[:thresh,0])
    fn[thresh-1] = numpy.sum(obs_quad_confusion_to_local_truth[:thresh,1])

fpr = fp/(fp+tn)
tpr = tp/(tp+fn)
precision = tp/(tp+fp)
recall = tp/(tp+fn)
accuracy = (tp+tn)/numpy.sum(obs_quad_confusion_to_local_truth)
print ' '.join(['%.3f' % x for x in accuracy])


f1_score = 2*precision*recall/(precision+recall)

print ' '.join(['%.3f' % x for x in f1_score])


# let's make a ROC curve, where we ask for varying numbers of experiments within a quad differing in phenotype from the 
# undrugged phenotype, how well we do compared to our "ground truth" model

tp = numpy.zeros((4,))
fp = numpy.zeros((4,))
tn = numpy.zeros((4,))
fn = numpy.zeros((4,))
for thresh in range(1,5):
    tp[thresh-1] = numpy.sum(quad_confusion_to_local_truth[thresh:,1])
    fp[thresh-1] = numpy.sum(quad_confusion_to_local_truth[thresh:,0])
    tn[thresh-1] = numpy.sum(quad_confusion_to_local_truth[:thresh,0])
    fn[thresh-1] = numpy.sum(quad_confusion_to_local_truth[:thresh,1])

fpr = fp/(fp+tn)
tpr = tp/(tp+fn)
precision = tp/(tp+fp)
recall = tp/(tp+fn)

f1_score = 2*precision*recall/(precision+recall)

print ' '.join(['%.3f' % x for x in f1_score])


# pylab.figure()
# pylab.plot(recall, precision, 'k.-'); 
# pylab.xlim(0,1)
# pylab.ylim(0,1)
# pylab.ylabel('precision')
# pylab.xlabel('recall')
# pylab.show()


myfpr = numpy.hstack([0, fpr[::-1], 1])
mytpr = numpy.hstack([0, tpr[::-1], 1])

auc = numpy.trapz(y=mytpr, x=myfpr)
pylab.title('AUC = '+"%.3f" % auc)
pylab.plot(myfpr, mytpr ,'k.-')
pylab.ylabel('TPR')
pylab.xlabel('FPR')
pylab.xlim(0,1)
pylab.ylim(0,1)
pylab.savefig('posthoc_results/20150222-'+PREFIX+'_local_quad_roc.svg')
pylab.clf()
pylab.close()



fout = open('THINGS_FOR_PAPER.txt','a')
quad_confusion_to_local_truth = quad_confusion_to_local_truth/numpy.sum(quad_confusion_to_local_truth,axis=1)[:,numpy.newaxis]
fout.write('local truth table'+"\n")
for i in range(5):
    fout.write(str(i)+' '+' '.join(str(x) for x in quad_confusion_to_local_truth[i,:])+"\n")

fout.write("\n")
fout.write('global truth table'+"\n")
quad_confusion_to_global_ground_truth = quad_confusion_to_global_ground_truth/numpy.sum(quad_confusion_to_global_ground_truth,axis=1)[:,numpy.newaxis]
for i in range(5):
    fout.write(str(i)+' '+' '.join(str(x) for x in quad_confusion_to_global_ground_truth[i,:])+"\n")

fout.close()


# make the excel spreadsheet of these summary data

clone_to_phenotype = {}
for x,ys in annotated_phenotypes.iteritems():
    for y in ys:
        clone_to_phenotype[y] = x


import xlwt

mergewb = xlwt.Workbook()
ws = mergewb.add_sheet('mean features')
ws.write(0, 0, 'Drug ID')
ws.write(0, 1, 'Clone ID')
ws.write(0, 2, 'Human-Assigned Class for Unperturbed Experiment')
ws.write(0, 3, 'Within-phenotype adjusted distance to unperturbed')
j = 4
for i in range(data.shape[1]):
    ws.write(0,j, 'z-scored SLF34.'+str(i+1))
    j += 1

ii = 1
for d in range(1,47):
    for pm1 in kosher_protein_list:
        ws.write(ii, 0, d)
        ws.write(ii, 1, pm1+1)
        ws.write(ii, 2, clone_to_phenotype[pm1+1])
        ws.write(ii, 3, unaugmented[unaugmented_clone_order.index(pm1+1),d-1])
        mvec = summary48[lookup48[(d,pm1+1)],:]
        j = 4
        for i in range(data.shape[1]):
            ws.write(ii,j, mvec[i])
            j += 1
        #
        ii += 1

mergewb.save('dataset-posthoc.xls')





# acc = []
# for d in range(1,48):
#     for p in range(1,49):
#         if p != 46:
#             acc.append((perturbations[kosher_protein_list.index(p-1),d-1], unaugmented[unaugmented_clone_order.index(p), d-1]))


# acc = numpy.asarray(acc)
# pylab.plot(acc[:,0], acc[:,1], 'k.')
# pylab.plot([threshold_for_global_ground_truth, threshold_for_global_ground_truth], [0, 1], 'b:')
# pylab.plot([0, numpy.max(acc[:,0])], [threshold_for_ground_truth, threshold_for_ground_truth], 'r:')

# from sklearn.metrics import confusion_matrix
# C = confusion_matrix(acc[:,0]>threshold_for_global_ground_truth, acc[:,1]>threshold_for_ground_truth)


# pylab.show()
import shutil

shutil.copy('posthoc_results/NORMALIZE_INNER_BY_MAX-augmented.svg', 'posthoc_results/Figure3.svg')