import numpy
import os
import pylab
import string
import itertools
import scipy
import scipy.stats
import scipy.spatial
import scipy.cluster
import scipy.cluster.hierarchy as sch

numpy.seterr(all='raise')


PREFIX = ''

NORMALIZE_INNER_BY_MAX = True

if NORMALIZE_INNER_BY_MAX:
    PREFIX += 'NORMALIZE_INNER_BY_MAX-'

#
# Helper functions
#

def linkage_to_control(Y):
    children_of = {}
    alive_at_level = [[]]
    for i in range(Y.shape[0]+1):
        children_of[i] = [i]
        alive_at_level[0].append(i)
    #
    for l in range(Y.shape[0]):
        alive_at_level.append([])
        i,j = Y[l,:2].astype('int')
        spot = Y.shape[0]+1+l
        children_of[spot] = children_of[i]+children_of[j]
        alive_at_level[l+1] = [x for x in alive_at_level[l] if x not in [i,j]]+[spot]
    #
    return children_of, alive_at_level

def myargsort(seq):
    # http://stackoverflow.com/questions/3382352/equivalent-of-numpy-argsort-in-basic-python/3383106#3383106
    #non-lambda version by Tony Veijalainen
    return [i for (v, i) in sorted((v, i) for (i, v) in enumerate(seq))]

#
# Load and arrange data
#

fin = open('metadata/drug_names.txt','r')
compound_labels = []
xx = 1
for line in fin:
    me = line.rstrip()
    if 'Alginate' not in line:
        me += 'mM'
    compound_labels.append(me)
    xx += 1

fin = open('metadata/clone_names.txt','r')
clone_labels = []
for line in fin:
    clone_labels.append(line.rstrip())

kosher_protein_list = [qwerty for qwerty in range(48) if clone_labels.index('NIH-3T3') != qwerty]

remapped_protein_subset = dict([(y+1,x) for (x,y) in enumerate(kosher_protein_list)])

modified_clone_labels = []
for z, i in enumerate(clone_labels):
    if 'Chr' in i[:3]:
        modified_clone_labels.append('clone '+str(z+1))
    else:
        modified_clone_labels.append(i)

modified_clone_labels = [modified_clone_labels[x] for x in kosher_protein_list]

enumerated_clone_labels = [y+' ('+str(x+1).zfill(2)+')' for (x,y) in zip(kosher_protein_list,modified_clone_labels)]

data = numpy.load('calculated/zscored_data_posthoc.npy')
theindices = numpy.load('calculated/indices_posthoc.npy')

ranges_of_48 = {}
for i in range(theindices.shape[0]):
    e = ( theindices[i,0]-1 ) % 48 + 1, ( theindices[i,1]-1 ) % 48 + 1
    if e not in ranges_of_48:
        ranges_of_48[e] = []
    ranges_of_48[e].append(i)

imputed_experiments = []
for e in itertools.product(range(1,49),range(1,49)):
    if e not in ranges_of_48:
        ranges_of_48[e] = []
        imputed_experiments.append(e)
        print 'imputing data for experiment',e

linkorder48 = ranges_of_48.keys()
lookup48 = dict(zip(linkorder48, range(len(linkorder48))))

summary48 = numpy.zeros((48*48,data.shape[1]))
for e,i in lookup48.iteritems():
    if e not in imputed_experiments:
        summary48[i,:] = numpy.mean(data[ranges_of_48[e],:],axis=0)

# impute to the unperturbed data
for e in imputed_experiments:
    summary48[lookup48[e],:] = summary48[lookup48[(48,e[1])],:]

#
# this is the raw matrix of how much each drugged experiment differed in feature space from the corresponding unperturbed experiment
#

protein_pooled_CV = numpy.zeros((47,))
for pp,p in enumerate(kosher_protein_list):
    acc = []
    for d in range(1,48):
        if (d,p+1) not in imputed_experiments:
            acc.extend(ranges_of_48[d,p+1])
    mydists = numpy.linalg.norm(data[acc]-summary48[lookup48[(48,p+1)],:],axis=1)
    protein_pooled_CV[pp] = numpy.std(mydists)/numpy.mean(mydists)

perturbations = numpy.zeros((47,47))
perturbations_CV = numpy.zeros((47,47))
perturbations_pooled_CV = numpy.zeros((47,47))
perturbations_imputed = numpy.zeros((47,47), dtype='bool')
for pp,p in enumerate(kosher_protein_list):
    for d in range(1,48):
        if (d,p+1) not in imputed_experiments:
            perturbations[pp,d-1] = numpy.sqrt(numpy.sum((summary48[lookup48[(d,p+1)],:]-summary48[lookup48[(48,p+1)],:])**2))
            dddd = numpy.mean(numpy.sqrt(numpy.sum((summary48[lookup48[(d,p+1)],:]-data[ranges_of_48[(d,p+1)],:])**2,axis=1)))
            if dddd > 0:
                perturbations_CV[pp,d-1] = perturbations[pp,d-1]/dddd
            else:
                perturbations_CV[pp,d-1] = perturbations[pp,d-1]
            perturbations_pooled_CV[pp,d-1] = perturbations[pp,d-1]/protein_pooled_CV[pp]
        else:
            perturbations_imputed[pp,d-1] = True





annotated_phenotypes = {}
xx = 1
for line in open('metadata/clone_assignments.txt','r'):
    ph = line.rstrip()
    if ph not in annotated_phenotypes:
        annotated_phenotypes[ph] = []
    annotated_phenotypes[ph].append(xx)
    xx += 1

for a in annotated_phenotypes.keys():
    b = annotated_phenotypes[a]
    border = numpy.argsort([clone_labels[x-1] for x in b])
    annotated_phenotypes[a] = [b[x] for x in border]



pheno_order = [
'nucleoli',
'nucleoli and cytoplasm',
'nucleoplasm', 
'nucleoplasm and cytoplasm',
'uniform cytoplasm',
'structured cytoplasm',
'focal',
'mitochondria',
'ER',
'ERGIC',
'Golgi',
'plasma membrane',
'vesicles',
'cytoskeleton']


basic_phenotype_groups = [[remapped_protein_subset[z] for z in annotated_phenotypes[x]] for x in pheno_order]
nbasic_phenotypes = len(basic_phenotype_groups)
groups_of_clones = [annotated_phenotypes[x] for x in pheno_order]

unaugmented_clone_order = reduce(lambda x,y: x+y,[[z for z in annotated_phenotypes[x]] for x in pheno_order])

#
# normalize the perturbations matrix by phenotype
#

inner_perturbations = numpy.zeros((len(basic_phenotype_groups),perturbations.shape[1]))
unaugmented = numpy.zeros(perturbations.shape)
unaugmented_imputed = numpy.zeros(perturbations.shape, dtype='bool')
augmented = numpy.zeros((len(basic_phenotype_groups)+perturbations.shape[0],perturbations.shape[1]))
augmented_imputed = numpy.zeros((len(basic_phenotype_groups)+perturbations.shape[0],perturbations.shape[1]), dtype='bool')
augmented_rownames = []
unaugmented_rownames = []
header_row = []
jj = 0
zz = 0
augmented_clone_remap = {}
unaugmented_phenotype_groups = []
for ii,ph in enumerate(basic_phenotype_groups):
    max_in_group = numpy.max(perturbations[ph,:])
    inner_perturbations[ii,:] = numpy.mean(perturbations[ph,:],axis=0)
    if NORMALIZE_INNER_BY_MAX:
        inner_perturbations[ii,:] /= max_in_group
        augmented[jj,:] = inner_perturbations[ii,:]
    else:
        augmented[jj,:] = inner_perturbations[ii,:]
    augmented_rownames.append(pheno_order[ii])
    header_row.append(True)
    jj += 1
    for kk in range(len(ph)):
        augmented_clone_remap[ph[kk]] = jj+kk
    if NORMALIZE_INNER_BY_MAX:
        augmented[jj:(jj+len(ph)),:] = perturbations[ph,:]/max_in_group
    else:
        augmented[jj:(jj+len(ph)),:] = perturbations[ph,:]
    unaugmented[zz:(zz+len(ph)),:] = augmented[jj:(jj+len(ph)),:]
    unaugmented_phenotype_groups.append(range(zz,len(ph)+zz))
    augmented_imputed[jj:(jj+len(ph)),:] = perturbations_imputed[ph,:]
    unaugmented_imputed[zz:(zz+len(ph)),:] = perturbations_imputed[ph,:]
    augmented_rownames.extend([modified_clone_labels[z] for z in ph])
    unaugmented_rownames.extend([modified_clone_labels[z] for z in ph])
    header_row.extend([False]*len(ph))
    jj += len(ph)
    zz += len(ph)

# numpy.save('posthoc_results/'+PREFIX+'unaugmented.npy', unaugmented)
# fout = open('posthoc_results/'+PREFIX+'unaugmented-rowdata.txt','w')
# for x,y in zip(unaugmented_clone_order, unaugmented_rownames):
#     fout.write(str(x)+' '+y+"\n")

# fout.close()


#
# Construct a clustering of the drugs
#

pdist_of_drugs = scipy.spatial.distance.pdist(inner_perturbations.T)

Z_drugs = scipy.cluster.hierarchy.linkage(pdist_of_drugs,method='average')
drugclus_children_of, drugclus_alive_at_level = linkage_to_control(Z_drugs)

CHOSEN_DRUG_CUTOFF = 30
# scipy.cluster.hierarchy.dendrogram(Z_drugs, labels=compound_labels, color_threshold=Z_drugs[CHOSEN_DRUG_CUTOFF,2])
# pylab.savefig('posthoc_results/'+PREFIX+'drug_clustering.svg', bbox_inches='tight')
# pylab.clf(); pylab.close()

# drug_summary = numpy.zeros((47,nbasic_phenotypes))
# for ph, phg in enumerate(basic_phenotype_groups):
#     d = 48
#     meanph = numpy.mean(numpy.vstack([summary48[lookup48[(d,p+1)],:] for p in phg if (d,p+1) in lookup48]),axis=0)
#     for d in range(1,48):
#         here = numpy.mean(numpy.vstack([summary48[lookup48[(d,p+1)],:] for p in phg if (d,p+1) in lookup48]),axis=0)
#         drug_summary[d-1,ph] = numpy.sqrt(numpy.sum((meanph-here)**2))

# pdist_of_stuff = scipy.spatial.distance.pdist(drug_summary)
# Z_drugs = scipy.cluster.hierarchy.linkage(pdist_of_stuff,method='average')
# drugclus_children_of, drugclus_alive_at_level = linkage_to_control(Z_drugs)

# CHOSEN_DRUG_CUTOFF = 38
# scipy.cluster.hierarchy.dendrogram(Z_drugs, labels=compound_labels, color_threshold=Z_drugs[CHOSEN_DRUG_CUTOFF,2])
# pylab.savefig('posthoc_results/'+PREFIX+'drug_clustering.svg', bbox_inches='tight')
# pylab.clf(); pylab.close()

drug_groups = [drugclus_children_of[x] for x in drugclus_alive_at_level[CHOSEN_DRUG_CUTOFF]]
# fout = open('posthoc_results/'+PREFIX+'groups_of_drugs.txt','w')
# for g in drug_groups:
#     fout.write(' '.join([str(x+1) for x in g])+"\n")

# fout.close()



exp_to_range = {}
for i in range(theindices.shape[0]):
    if (theindices[i,0], theindices[i,1]) not in exp_to_range:
        exp_to_range[(theindices[i,0], theindices[i,1])] = []
    exp_to_range[(theindices[i,0], theindices[i,1])].append(i)

#linkorder96_subset = [x for x in exp_to_range.keys() if ( (x[1] - 1) % 48 in kosher_protein_list ) and (48,x[1]) in exp_to_range]

linkorder96_subset = [x for x in exp_to_range.keys() if ( (x[1] - 1) % 48 +1 in unaugmented_clone_order ) and (48,x[1]) in exp_to_range]

linkorder96_subset_set = set(linkorder96_subset)

ap = open('sample/bear_round30.apredictions','r')
exp_to_pheno = {}
all_exp_to_pheno = {}
pheno_to_exp = {}
for rline in ap:
    Line = string.split(rline.rstrip())
    e = (int(Line[0]), int(Line[1]))
    all_exp_to_pheno[e] = int(Line[2])
    if e in linkorder96_subset_set:
        exp_to_pheno[e] = int(Line[2])
        if int(Line[2]) not in pheno_to_exp:
            pheno_to_exp[int(Line[2])] = []
        pheno_to_exp[int(Line[2])].append(e)


linkorder96_map = dict(zip(linkorder96_subset, range(len(linkorder96_subset))))

summary96 = numpy.zeros((len(linkorder96_subset),data.shape[1]))
for i,e in enumerate(linkorder96_subset):
    summary96[i,:] = numpy.mean(data[exp_to_range[e],:],axis=0)


pheno_to_representation = {}
for ph,es in pheno_to_exp.iteritems():
    places = []
    for e in [x for x in es if x in linkorder96_subset_set]:
        places.extend(exp_to_range[e])
    pheno_to_representation[ph] = numpy.mean(data[places,:],axis=0)

drugged_96 = [x for x in linkorder96_subset if (x[0] not in [48,96]) and exp_to_pheno[x] in pheno_to_representation]


helloacc = []
sameacc = []
wacc = []
helloacc_per_pheno = {}
for x in linkorder96_subset:
    if x[0] != 48 and x[0] != 96 and ( ( x[1] -1 ) % 48 ) in kosher_protein_list: # and ( ( x[1] -1 ) % 48 ) in spreads:
        if exp_to_pheno[x] != exp_to_pheno[(48,x[1])]:
            meas = numpy.sqrt(numpy.sum((summary96[linkorder96_map[x]]-summary96[linkorder96_map[(48, x[1])]])**2))
            predat = numpy.sqrt(numpy.sum((pheno_to_representation[exp_to_pheno[x]]-summary96[linkorder96_map[(48, x[1])]])**2))
            helloacc.append([meas, predat])
            wacc.append(x)
            helloacc_per_pheno.setdefault(exp_to_pheno[x],[]).append([meas, predat])
        else:
            meas = numpy.sqrt(numpy.sum((summary96[linkorder96_map[x]]-summary96[linkorder96_map[(48, x[1])]])**2))
            predat = numpy.sqrt(numpy.sum((pheno_to_representation[exp_to_pheno[x]]-summary96[linkorder96_map[(48, x[1])]])**2))
            sameacc.append([meas, predat])


sameacc = numpy.asarray(sameacc)

helloacc = numpy.asarray(helloacc)

fout = open('THINGS_FOR_PAPER_post_reviewer.txt','w')

fout.write('median difference='+str(numpy.median(sameacc,axis=0)-numpy.median(helloacc,axis=0))+"\n")
fout.write('mm mann-whitney='+str(scipy.stats.mannwhitneyu(sameacc[:,0],helloacc[:,0]))+"\n")
fout.write('pp mann-whitney='+str(scipy.stats.mannwhitneyu(sameacc[:,1],helloacc[:,1]))+"\n")


fout.write('mm pearson r='+str(scipy.stats.pearsonr(sameacc[:,0], sameacc[:,1]))+"\n")
fout.write('pp pearson r='+str(scipy.stats.pearsonr(helloacc[:,0], helloacc[:,1]))+"\n")

fout.close()

#
# only via unperturbed
#

# clone_stds = numpy.zeros((48,))
# for p in numpy.asarray(kosher_protein_list)+1:
#     vs = []
#     ns = []
#     for pp in [0,48]:
#         for d in [48, 96]: #range(1,97):
#             e = d,p+pp
#             if e in linkorder96_subset:
#                 vs.append(numpy.mean(numpy.sum((data[exp_to_range[e],:]-summary96[linkorder96_subset.index(e)])**2,axis=1)))
#                 ns.append(len(exp_to_range[e]))
#     vs = numpy.asarray(vs)
#     ns = numpy.asarray(ns)
#     if len(vs)>1:
#         clone_stds[p-1] = numpy.sqrt(numpy.sum((ns-1)*vs)/(numpy.sum(ns)-2))
#     elif len(vs)==1:
#         clone_stds[p-1] = numpy.sqrt(vs[0])
#     else:
#         print 1/0

# helloacc = []
# for x in linkorder96_subset:
#     x48 = (x[0]-1) % 48, (x[1]-1) % 48
#     if x48 not in imputed_experiments and x48[1] in kosher_protein_list and x48[0] < 47:
#         # meas = perturbations[x48[0], kosher_protein_list.index(x48[1])]/clone_stds[x48[1]]
#         meas = perturbations_pooled_CV[x48[0], kosher_protein_list.index(x48[1])]
#         # obs = numpy.sqrt(numpy.sum((pheno_to_representation[exp_to_pheno[x]]-summary48[lookup48[(48,x48[0]+1)],:])**2))/clone_stds[x48[1]]
#         obs = numpy.sqrt(numpy.sum((pheno_to_representation[exp_to_pheno[x]]-summary48[lookup48[(48,x48[0]+1)],:])**2))/clone_stds[x48[1]]
#         helloacc.append([meas, obs])

# helloacc = numpy.asarray(helloacc)

# scipy.stats.pearsonr(helloacc[:,0], helloacc[:,1])







